"""
Station Class
"""
# Standard library modules
import logging

# Non-standard modules
from obspy.core.inventory.station import Station as obspy_Station

from obspy.core.inventory.util import (Comment)
from obspy.core.utcdatetime import UTCDateTime
# from obspy.taup.seismic_phase import self_tokenizer

# obsinfo modules
from .processing import Processing
from ..instrumentation import Instrumentation, Instrumentations
from ..helpers import (Location, Locations, OIDate,
                       str_list_str, str_indent, verify_dict_is_empty,
                       ObsinfoClassList, Comments, ExternalReferences, Identifiers)
from .operator import Operators
from ..obsmetadata import ObsMetadata
from .site import Site

logger = logging.getLogger("obsinfo")


class Stations(ObsinfoClassList):
    """
    A list of Station objects
    """
    def __init__(self, stations_dict, station_only, operators, comments):
        """
        Args:
            stations_dict: (dict or :class:`.ObsMetadata`): dictionary
                from station or network info file with YAML or JSON
                attributes
            station_only (boolean): Creates object with no instrumentation
            operators (:class:`.Operators`): default station
                operators
            comments: (:class:`.Comments`): default station comments
        """
        if stations_dict is None:
            super().__init__([])
        else:
            super().__init__([Station(k, v, station_only, operators, comments)
                             for k, v in (stations_dict.items())], Station)


class Station(object):
    """
    Station. Equivalent to obspy/StationXML Station

    Methods convert info files to an instance of this class and convert the
    object to an `obspy` object.

    Attributes:
        code (str): Station code (FDSN)
        site (str): Site name
        start_date (str with date format): station start date
        end_date (str with date format): station end date
        location_code (str):
        restricted_status (str): status of station
        locations (list of :class:`.Location`)
        location (:class:`.Location`): default location code
            of channels, corresònding to `location_code`
        instrumentation (:class:`.Instrumentation` or list of
            :class:`.Instrumentation`):
        processing (list of objects of :class:`.Processing`) : attributes
            for clock correction processing
        comments (list of str):
        extras (list of str):
        source_id (str): StationXML sourceID (URI format)
        identifiers: (list of str): permanent identifiers (URI format)
        extra_references (list of dict): external references, each one is
            a dict with keys ['uri', 'description']
        water_level (number): elevation (m) of water surface (useful for lakes)
        obspy_station (:class:`obspy.core.inventory.station.Station`):
            Equivalent obspy object
    """

    def __init__(self, code, attributes_dict, station_only=False,
                 stations_operators=None, stations_comments=None):
        """
        Constructor

        Args:
            attributes_dict: (dict or :class:`.ObsMetadata`): dictionary
                from station or network info file with YAML or JSON attributes
            station_only (boolean): Creates object with no instrumentation
            stations_operators (:class:`.Operators`): default station
                operator(s)
            stations_comments: (:class:`.Comments`): default station comments
        Raises:
            TypeError
        """
        if not attributes_dict:
            msg = 'No station attributes'
            logger.error(msg)
            raise TypeError(msg)

        logger.info(f'Creating Station "{code}"')
        self.code = code
        self.site = Site(attributes_dict.pop("site", None))
        self.start_date = OIDate(attributes_dict.pop("start_date"))
        self.end_date = OIDate(attributes_dict.pop("end_date"))
        self.location_code = attributes_dict.pop("location_code")
        self.restricted_status = attributes_dict.pop("restricted_status", None)
        self.operators = Operators(attributes_dict.pop("operators", None))
        self.description = attributes_dict.pop("description", None)

        self.locations = Locations.from_locations_dict(attributes_dict.pop('locations'))
        self.location = self.locations.get_by_loc_code(self.location_code)

        if 'instrumentations' in attributes_dict:
            instr_dict_list = attributes_dict.pop('instrumentations', None)
            self.instrumentations = Instrumentations(
                [Instrumentation(instr_dict, self.locations,
                                 start_date, end_date)
                 for instr_dict in instr_list])
        else:
            instr_dict = attributes_dict.pop('instrumentation', None)
            if isinstance(instr_dict, str) or instr_dict is None:
                self.instrumentations = Instrumentations([])
            else:
                self.instrumentations = Instrumentations([Instrumentation(
                    instr_dict, self.locations, self.location_code,
                    self.start_date.date, self.end_date.date)])

        self.processing = Processing(attributes_dict.pop('processing', None))
        self.comments = Comments(attributes_dict.pop("comments", []))
        if stations_comments is not None:
            self.comments += stations_comments
        self.source_id = attributes_dict.pop('source_id', None)
        self.identifiers = Identifiers(attributes_dict.pop('identifiers', None))
        self.extras = attributes_dict.pop('extras', None)
        self.external_references = ExternalReferences(attributes_dict.pop('external_references', None))
        self.comments += Comments.from_extras(attributes_dict.pop('extras', None))
        self.comments += self.processing.to_comments()
        if 'notes' in attributes_dict:
            del attributes_dict['notes']
        verify_dict_is_empty(attributes_dict)

    def __repr__(self):
        s =  f'station(dict(code={self.code}, site={self.site}, '
        s += f'start_date={self.start_date}, end_date={self.end_date}, '
        s += f'location_code={self.location_code}, '
        s += f'{len(self.locations)} Locations, '
        if self.processing:
            s += f'processing-steps: {self.processing.processing_list}'
        s += ')'
        return s

    def __str__(self, indent=0, n_subclasses=0):
        if n_subclasses < 0:
            return f'{self.__class__.__name__}: {self.code}'
        kwargs = dict(indent=4, n_subclasses=n_subclasses-1)
        s = f'{self.__class__.__name__}:\n'
        s += f'    code: {self.code}\n'
        s += f'    site: {self.site}\n'
        s += f'    start_date: {self.start_date}\n'
        s += f'    end_date: {self.end_date}\n'
        s += f'    location_code: {self.location_code}\n'
        s += f'    restricted_status: {self.restricted_status}\n'
        s += f'    locations: {self.locations.__str__(**kwargs)}\n'
        s += f'    locations: {self.location.__str__(**kwargs)}\n'
        s += f'    instrumentations: {self.instrumentations.__str__(**kwargs)}\n'
        s += f'    processing: {self.processing}'
        if len(self.comments) > 0:
            s += '\n    comments: {str_list_str(self.comments, **kwargs}'
        if len(self.external_references) > 0:
            s += '\n    external_references: {str_list_str(self.external_references, **kwargs}'
        if self.source_id is not None:
            s += '\n    source_id: {self.source_id}'
        if self.description is not None:
            s += '\n    description: {self.description}'
        if len(self.identifiers) > 0:
            s += '\n    identifiers: {str_list_str(self.identifiers,  **kwargs)}'
        if self.extras is not None:
            s += '\n    extras: {str_list_str(self.extras, **kwargs)}'
        return str_indent(s, indent)

    def to_obspy(self):
        """
        Convert station object to obspy object

        Returns:
            (:class:`obspy.core.inventory.station.Station`):
                  corresponding obspy Station object
        """
        channels_number = 0
        chnl_list = []
        equip_list = []
        for x in self.instrumentations:
            channels_number += len(x.channels)
            chnl_list += [ch.to_obspy() for ch in x.channels]
            equip_list += [x.equipment]

        if len(self.external_references) > 0:
            msg = ('obspy 1.4.0 could not create station-level '
                   "external_references, so your's will not be written. "
                   "If this has been fixed in obspy, put an issue on the "
                   "obsinfo gitlab page so that we can update this")
            logger.warning(msg)

        return obspy_Station(
            code=self.code,
            latitude=self.location.latitude.to_obspy(),
            longitude=self.location.longitude.to_obspy(),
            elevation=self.location.elevation.to_obspy(),
            water_level=self.location.water_level,
            channels=chnl_list,
            site=self.site.to_obspy(),
            vault=self.location.vault,
            geology=self.location.geology,
            equipments=equip_list,
            operators=self.operators.to_obspy(),
            creation_date=self.start_date.to_obspy(),
            termination_date=self.end_date.to_obspy(),
            # external_references=self.external_references.to_obspy(),
            total_number_of_channels=channels_number,
            selected_number_of_channels=channels_number,
            comments=self.comments.to_obspy(),
            start_date=self.start_date.to_obspy(),
            end_date=self.end_date.to_obspy(),
            restricted_status=self.restricted_status,
            source_id=self.source_id,
            identifiers=self.identifiers.to_obspy(),
            description=self.description,
            alternate_code=None,
            historical_code=None,
            data_availability=None)

    def add_extras_to_comments(self):
        """
        Convert processing info and extras to comments
        """

        if self.extras:
            self.comments.append('EXTRA ATTRIBUTES (for documentation only):')
            self.comments = self.comments + self.extras
        if self.processing.processing_list:
            self.comments.append(self.processing.processing_list)



"""
Python equivalent of obspy FloatWithUncertaintiesFixedUnit class
"""
# Standard library modules
import warnings
import logging

# Non-standard modules
from obspy.core.util.obspy_types import (FloatWithUncertaintiesFixedUnit,
                                         FloatWithUncertaintiesAndUnit)


warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class FloatWithUncert(object):
    """
    Python equivalent of obspy FloatWithUncertaintiesFixedUnit and
    FloatWithUncertaintiesAndUnit

    Attributes:
        value (float): float value
        uncertainty (float): uncertainty in value
        measurement_method (str): measurement method
    """

    def __init__(self, attributes_dict):
        """
        Create object and assign attributes from attributes_dict.

        Args:
            attributes_dict (dict or :class:`ObsMetadata`): dict with
                relevangt keys
        """
        self.value = attributes_dict['value']        # required
        self.uncertainty = attributes_dict.get('uncertainty', None)
        self.measurement_method = attributes_dict.get('measurement_method',
                                                      None)
        self.unit = attributes_dict.get('unit', None)

    def __str__(self, indent=0, n_subclasses=0):
        """ Writes everything out, one line, no subclasses """
        s = 'FloatWithUncert: {} +- {}'.format(self.value, self.uncertainty)
        if self.unit is not None:
            s += f' {self.unit}'
        if self.measurement_method is not None:
            s += f', measurement_method = {self.measurement_method}'
        return s

    def __repr__(self, no_title=False):
        """
        Args:
            no_title (bool): don't surround dict with 'FloatWithUncert()'
        """
        args = [f"'value': {self.value}"]
        if self.uncertainty:
            args.append(f"'uncertainty': {self.uncertainty}")
        if self.measurement_method:
            args.append(f"'measurement_method': '{self.measurement_method}'")
        if self.unit:
            args.append(f"'unit': '{self.unit}'")
        dict_string = '{' + ", ".join(args) + '}'
        if no_title:
            return dict_string
        return 'FloatWithUncert(' + dict_string + ')'

    def to_obspy(self):
        """
        Return obspy object:
          - FloatWithUncertaintiesFixedUnit if unit=None
          - FloatWithUncertaintiesAndUnit otherwise
        """
        if self.unit is None:
            return FloatWithUncertaintiesFixedUnit(
                value=self.value,
                lower_uncertainty=self.uncertainty,
                upper_uncertainty=self.uncertainty,
                measurement_method=self.measurement_method)
        else:
            return FloatWithUncertaintiesAndUnit(
                value=self.value,
                lower_uncertainty=self.uncertainty,
                upper_uncertainty=self.uncertainty,
                measurement_method=self.measurement_method,
                unit=self.unit)



import xml.etree.ElementTree as ET

error = lambda s: "\033[91m" + s + "\033[0m"


class XmlTree:

    ns = "{http://www.fdsn.org/xml/station/1}"

    @staticmethod
    def convert_string_to_tree(xmlString):

        return ET.fromstring(xmlString)

    @staticmethod
    def getroot(xml):
        return xml.getroot()

    @staticmethod
    def add_ns(tag):
        return f"{XmlTree.ns}{tag}"

    def xml_compare(self, x1, x2, excludes=["Created"], excludes_attributes=[]):
        """
        Compares two xml etrees
        :param x1: the first tree
        :param x2: the second tree
        :param excludes: list of string of attributes to exclude from comparison
        :return:
            True if both files match
        """
        # get namespace

        if x1.tag != x2.tag:
            return False, error(f"Tags do not match: {x1.tag} and {x2.tag}")
        for name, value in x1.attrib.items():
            if not name in excludes_attributes:
                if x2.attrib.get(name) != value:
                    print(
                        error(
                            f"Attributes do not match: {name}={value}, {name}={x2.attrib.get(name)}"
                        )
                    )
                    print(error(f"Elements: {x1.attrib},{x2.attrib}"))
                    return False
        for name in x2.attrib.keys():
            if not name in excludes:
                if name not in x1.attrib:
                    print("x2 has an attribute x1 is missing: %s" % name)

                    return False
        if not self.text_compare(x1.text, x2.text):
            print(error(f"text: {x1.text} != {x2.text}"))
            print(error(f"Elements: {x1.tag},{x2.tag}"))
            return False
        if not self.text_compare(x1.tail, x2.tail):
            print("tail: %r != %r" % (x1.tail, x2.tail))
            return False
        cl1 = list(x1)
        cl2 = list(x2)

        if len(cl1) != len(cl2):
            print(error(f"children length differs, {len(cl1)}!= {len(cl2)}"))
            return False
        i = 0
        for c1, c2 in zip(cl1, cl2):
            i += 1
            if not c1.tag in excludes:
                if not self.xml_compare(c1, c2, excludes,excludes_attributes):
                    print(c1.attrib,c2.attrib)
                    print(error(f"children {c1.tag} do not match with {c1.tag}"))
                    return False
        return True

    def text_compare(self, t1, t2):
        """
        Compare two text strings
        :param t1: text one
        :param t2: text two
        :return:
            True if a match
        """
        if not t1 and not t2:
            return True
        if t1 == "*" or t2 == "*":
            return True
        return (t1 or "").strip() == (t2 or "").strip()


def main():
    print("Should not be run alone")

if __name__ == "__main__":
    main()



# import os
# import sys
# import re
from pathlib import Path
import warnings
import yaml
import logging

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Singleton:
    """
    Class to implement singleton pattern design in Python
    """

    def __init__(self, cls):
        self._cls = cls

    def Instance(self):
        try:
            return self._instance
        except AttributeError:
            self._instance = self._cls()
            return self._instance

    def __call__(self):
        msg = 'Singletons must be accessed through `Instance()`.'
        logger.debug("Programming error: " + msg)
        raise AttributeError(msg)

    def __instancecheck__(self, inst):
        return isinstance(inst, self._cls)


@Singleton
class ObsinfoConfiguration:
    """
    Singleton class to store obsinfo configuration, basically its directories to store files
    and its version

    Attributes:
        obsinfo_datapath: list of strings representing directories to search for info files
        gitlab_repository (str): remote Gitlab repository name
        gitlab_project (str): project within the repository
        gitlab_path (str): path to info files within the project
        obsinfo_version (str): version of this program
    """

    logger = None

    def __init__(self):
        """
        Read .obsinforc file in home directory and setup the configuration
        :raises: FileNotFoundError, OSError
        """

        home = Path.home()
        config_file = Path(home).joinpath(".obsinforc")
        # This is called directly to avoid the paraphernalia of yamlref.
        try:
            with open(config_file, "r") as fp:
                config_dict = yaml.safe_load(fp)
        except FileNotFoundError:
            msg = "Configuration file .obsinforc not found. Run obsinfo-setup"
            print(msg)
            logger.error(msg)
            raise
        except OSError:
            msg = "Operating system error reading configuration file .obsinforc"
            print(msg)
            logger.error(msg)
            raise

        self.obsinfo_datapath = config_dict.get("datapath", "")
        self.gitlab_repository = config_dict.get("gitlab_repository", "")
        self.gitlab_project = config_dict.get("gitlab_project", "")
        self.gitlab_path = config_dict.get("gitlab_path", "")
        self.obsinfo_branch = str(config_dict.get("obsinfo_branch", ""))



"""
Phone class
"""
# Standard library modules
import warnings
import re
import logging
from string import digits

from obspy.core.inventory.util import PhoneNumber

from .functions import str_indent
from .obsinfo_class_list import ObsinfoClassList

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Phones(ObsinfoClassList):
    """
    List of Phone objects
    """
    def __init__(self, seq):
        """
        Args:
            seq: (list of Phone or None): list of attribute dicts
                    describing a Phone
        """
        seq = [Phone.str_to_dict(x) if isinstance(x, str) else x for x in seq]
        super().__init__(seq, Phone)


class Phone(object):
    """
    Equivalent of StationXMl PhoneNumber

    Attributes:
        description (str): Description of the phone # (work, mobile...)
        country_code (str): Telephone country code
        area_code (int): Telephone area code
        phone_number (str): Telephone number
    """

    def __init__(self, attributes_dict_or_str):
        """
        Phone # must be provided as one of the following:
            - an international format str:
                "+{country_code}{area_code}{phone_number}"
                only numbers are allowed, no dashes, no parentheses, etc
            - a US format string (for legacy):
                "+{country_code} ({area_code}) {phone_number}"
                "00{country_code} ({area_code}) {phone_number}"
                where phone_number can have dashes
            - a dict with fields "country_code, "area_code" and "phone_number"
        Args:
            attributes_dict_or_str (dict, :class:`.ObsMetadata`, or str):
                phone information
        """
        if not attributes_dict_or_str:
            raise ValueError('No str or attributes dict!')

        if isinstance(attributes_dict_or_str, str):
            # self.convert_phone_number(attributes_dict_or_str)
            attributes_dict_or_str = self.str_to_dict(attributes_dict_or_str)
        self.phone_number = attributes_dict_or_str['phone_number']
        self.description = attributes_dict_or_str.get('description', None)
        self.country_code = attributes_dict_or_str.get('country_code',
                                                           None)
        self.area_code = int(attributes_dict_or_str.get('area_code', 0))
        self._clean_phone_number()

    def __str__(self, indent=0, n_subclasses=0):
        """Always a one_liner"""
        s = f"{self.__class__.__name__}: "
        if self.country_code:
            s += f'+{self.country_code} '
        if not self.area_code == 0:
            s += f'{self.area_code}'
        s += f'{self.phone_number}'
        if self.description:
            s += ' f({self.description})'
        return str_indent(s, indent)

    def __repr__(self):
        args = []
        if self.country_code:
            args.append(f'country_code="{self.country_code}"')
        if self.area_code:
            args.append(f'area_code={self.area_code}')
        args.append(f'phone_number="{self.phone_number}"')
        if self.description:
            args.append(f'description="{self.description}"')
        return 'Phone(dict(' + ', '.join(args) + '))'

    @staticmethod
    def str_to_dict(phone):
        """
        Return attribute_dict corresponding to phone number string
        Try to convert international numbers to the FDSN American standard.
        If already in American standard, use area code.
        Requires country codes, because we're not just the USA!

        Args:
            phone (str): phone number in (hopefully) one of several
                recognisable formats
        Returns:
            (tuple):
                ``country_code``
                ``area_code``: default=0
                ``phone_number``
                
        >>> Phone("+33 6 12345678")
        Phone(dict(country_code="33", phone_number="612345678"))
        >>> str(Phone("+33 6 12345678"))
        '+33 612345678'
        >>> Phone("+33 (6) 12345678")
        Phone(dict(country_code="33", phone_number="612345678"))
        >>> Phone("+1 (415) 123-4567")
        Phone(dict(country_code="1", area_code=415, phone_number="1234567"))
        >>> Phone("001 (415) 123-4567")
        Phone(dict(country_code="1", area_code=415, phone_number="1234567"))
        >>> Phone("+001 (415) 123-4567")
        Phone(dict(country_code="1", area_code=415, phone_number="1234567"))
        >>> Phone("(415) 123-4567")
        ValueError: "(415) 123-4567" returned no phone number
        """
        attr_dict = {'area_code': 0,
                     'phone_number': "",
                     'country_code': None}

        # For reference:
        # country = re.compile("^(\+{0,1}|00)[0-9]{1,3}$")
        # area = re.compile("^\({0,1}[0-9]{3}\){0,1}$")
        # phone = re.compile("^[0-9]{3}\-[0-9]{4}$")

        us_phone_match = re.match('(\\+00|\\+|00)(?P<country>[0-9]{1,3}) '
                                  '*(?P<area>\\({0,1}[0-9]{3}\\){0,1}) '
                                  '*(?P<phone>[0-9]{3}\\-[0-9]{4})$',
                                  phone)

        if us_phone_match:
            # print(us_phone_match.groupdict())
            attr_dict['country_code'] = us_phone_match.group('country')
            attr_dict['area_code'] = int(''.join(c for c in us_phone_match.group('area')
                                         if c in digits))
            attr_dict['phone_number'] = ''.join(c for c in us_phone_match.group('phone')
                                        if c in digits)
        else:
            c_code_plus_ptn = '^\+([0-9]{1,3})'
            c_code_zero_ptn = "^00([0-9]{1,3})"
            c_code_plus_re = re.compile(c_code_plus_ptn)
            c_code_zero_re = re.compile(c_code_zero_ptn)
            phone_ptn = "(?P<phone>(\([0-9]+\))* *([0-9]+[ \-\.]*)*[0-9]+)$"

            c_code = c_code_plus_re.match(phone)
            if not c_code:  # | for alternatives in regex doesn't work
                c_code = c_code_zero_re.match(phone)
                phone_re = re.compile(c_code_zero_ptn + " *" + phone_ptn)
            else:
                phone_re = re.compile(c_code_plus_ptn + " *" + phone_ptn)

            if c_code:
                attr_dict['country_code'] = c_code.group(1)

            mtch = phone_re.match(phone)
            if mtch:
                phone_number = mtch.group('phone')
                # The following is done to avoid FDSN reg exp restrictions
                # for phones, American based
                for chr in ["(", ")", ".", "-", " "]:
                    phone_number = phone_number.replace(chr, "")
                # self.phone_number = phone_number[0:3] + "-" + phone_number[3:]
                attr_dict['phone_number'] = phone_number
        if not attr_dict['phone_number']:
            raise ValueError(f'"{phone}" returned no phone number')
        return attr_dict

    def _clean_phone_number(self):
        """
        Make sure there are no non-digits
        """
        self.phone_number = "".join([c for c in self.phone_number if c in digits])
        if self.country_code is not None:
            self.country_code = "".join([c for c in self.country_code if c in digits])

    def to_obspy(self):
        """ Stupidly, phone_number has to have a dash inside of it 
        FDSN pattern search rule, copied by obspy
        """
        return PhoneNumber(self.area_code,
                           self.phone_number[:3] + '-' + self.phone_number[3:],
                           self.country_code,
                           self.description)



"""
ExternalReferences Class, used in Channels and Stations
"""
# Standard library modules
import warnings
import logging
import json

from obspy.core.inventory.util import ExternalReference as obspy_extref

from .functions import str_indent
from .person import Persons
from .oi_date import OIDate
from .obsinfo_class_list import ObsinfoClassList

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class ExternalReferences(ObsinfoClassList):
    """
    A list of ExternalReference objects
    """
    def __init__(self, attributes_list):
        """
            Args:
                attributes_list: (list of dict or str): list
                    of external references read from YAML/JSON file
        """
        if attributes_list is None:
            super().__init__([], ExternalReference)
        else:
            super().__init__(attributes_list, ExternalReference)


class ExternalReference(object):
    def __init__(self, attributes_dict):
        self.uri = attributes_dict.pop('uri')
        self.description = attributes_dict.pop('description')

    def __str__(self, indent=0, n_subclasses=0):
        if n_subclasses < 0:
            return(f'{self.__class__.__name__}: {self.uri}')
        s = f'{self.__class__.__name__}:\n'
        s += f'    uri: {self.uri}\n'
        s += f'    description: {self.description}'
        return str_indent(s, indent)

    def to_obspy(self):
        return obspy_extref(uri=self.uri, description=self.description)



"""
ObsinfoClassList template class

Use as a template for all lists of obsinfo classes.
Defines how __str__ is to be printed and checks to make sure that
all elements are of the desired class
"""
import warnings
import logging
import inspect

from .functions import str_indent

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class ObsinfoClassList(list):
    def __init__(self, seq, element_class):
        """
        Args:
            seq (list or None): sequence of dicts or of :class:element_class.
                If dicts, converts to element_class type using element_class()
                If None, return empty instance
            element_class (:class:): the class that every element should be
                (or become)
        """
        if seq is not None and not isinstance(seq, list):
            raise TypeError(f'seq ({type(seq)}) is neither a list nor None')
        # Validate element_class
        if not inspect.isclass(element_class):
            raise TypeError('element_class is not a class')
        self.element_class = element_class
        
        if seq is None:
            super().__init__([])
            return
        elif len(seq) == 0:
            super().__init__([])
            return
        
        # Verify all elements are of same type
        for x in seq:
            if not isinstance(x, type(seq[0])):
                raise TypeError(f'seq elements are not all of same type '
                                f'({type(x)} != {type(seq[0])})')

        if not isinstance (seq[0], element_class):
            seq = [element_class(x) for x in seq]

        super().__init__(x for x in seq)

    def append(self, item):
        if not isinstance(item, self.element_class):
            item = element_class(item)
        super().append(item)

    def __str__(self, indent=0, n_subclasses=0):
        s = f'{self.__class__.__name__}:'
        if len(self) == 0:
            return s + ' []'
        elif len(self) == 1:
            return s + f' [{self[0].__str__(4, n_subclasses-1)}]'
        if n_subclasses < 0:
            return s + f' {len(self)} {self.element_class}s'
        for x in self:
            s += f'\n    - {x.__str__(indent=8, n_subclasses=n_subclasses-1)}'
        return str_indent(s, indent)
        
    def to_obspy(self):
        """Return list of element.to_obspy()"""
        if len(self) == 0:
            return None
        if getattr(self[0], 'to_obspy'):
            return [x.to_obspy() for x in self]
        else:
            raise ValueError('{self.element_class.__name__} class has no to_obspy() method')



"""
Exit values as constants as per UNIX BSD standard
"""

# Succesful treatment
EXIT_SUCCESS = 0
# File could not be processed, generic reasons
EXIT_FAILURE = 1

# Reference: https://www.freebsd.org/cgi/man.cgi?query=sysexits&manpath=FreeBSD+4.3-RELEASE

# bad CLI parameters
EXIT_USAGE = 64
# bad file content (input file : Network metadata file)
EXIT_DATAERR = 65
# file not found (input file : Network metadata file)
EXIT_NOINPUT = 66
# A service is unavailable (This can occur if a support program or file does not exist)
EXIT_UNAVAILABLE = 69

# (technical/internal error)
# An internal software error has been detected
EXIT_SOFTWARE = 70

# Can't create file (report file, output file)
EXIT_CANTCREAT = 73
# An error occurred while doing I/O on some file */
EXIT_IOERR = 74

# Something was found in an unconfigured or misconfigured state */
EXIT_CONFIG = 78



"""
Instrumentation class
"""
# Standard library modules
import warnings
import logging

# Non-standard modules

# obsinfo modules
from ..obsmetadata import ObsMetadata
from .instrument_component import Equipment
from .channel import Channel, Channels
from ..helpers import str_indent, verify_dict_is_empty, Location, ObsinfoClassList

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Instrumentations(ObsinfoClassList):
    """
    A list of Instrumentation objects
    """
    def __init__(self, instrumentations_list):
        """
        Args:
            instrumentations_list: (list of :class:`Instrumentations`):
        """
        if instrumentations_list is None:
            super().__init__([], Instrumentation)
        elif isinstance(instrumentations_list, list):
            super().__init__(instrumentations_list, Instrumentation)
        else:
            raise TypeError('instrumentation_list is neither a list nor None')


class Instrumentation(object):
    """
    One or more Instrument Channels. Part of an obspy/StationXML Station

    Methods convert info files to an instance of this class. No equivalent
    obspy/StationXML class

    A more detailed description the class and its attributes is found in XXX

    Attributes:
        equipment (:class:`Equipment`):
        channels (list): list of channels (:class:`Channel`)
    """
    def __init__(self, attributes_dict, station_locations, station_location_code,
                 station_start_date, station_end_date):
        """
        Constructor

        attributes_dict may contain a configuration_selection for the
        instrumentation and the corresponding configs for the components:
        datalogger, preamplifier and sensor

        Args:
            attributes_dict (dict or :class:`ObsMetadata`):
                instrumentation attributes
            station_locations (:class:`Locations`):  list of Locations
            station_location_code (str): station's location code
            station_start_date (str): station start date
            station_end_date (str): station end date

        It is assumed an instrumentation's default location,
        start date and end_date are the same as its station's.
        """
        if not isinstance(station_location_code, str):
            raise TypeError(f'station_location_code is a {type(station_location_code)}, should be a str')
        if isinstance(attributes_dict['base'], str):
            # No base instrument defined, should act like a None
            self.channels = Channels()
            self.equipment = Equipment({'description': attributes_dict['base']})
            return
        
        ic_names = ('datalogger', 'sensor', 'preamplifier')

        self.input = {'attributes_dict': {} if not attributes_dict else '<ObsMetaData>',
                      'station_locations': station_locations,
                      'station_start_date': station_start_date,
                      'station_end_date': station_end_date,
                      'station_location_code': station_location_code}
        
        # Syntax checking - Check whether
        if not attributes_dict:
            msg = 'No instrumentation attributes'
            warnings.warn(msg)
            logger.error(msg)
            raise TypeError(msg)
        elif isinstance(attributes_dict, dict):
            if not isinstance(attributes_dict, ObsMetadata):
                attributes_dict = ObsMetadata(attributes_dict)
        else:
            # Easy mistake of including a dash in the yaml file
            msg = 'Instrumentation is not a dict'
            logger.error(msg)
            raise TypeError(msg)

        # Remove elements not to be processed here
        channel_modifs = attributes_dict.pop('channel_modifications', ObsMetadata({}))

        # Extract instrument_component modifications
        mods = attributes_dict.get('modifications', None)
        if mods is not None:
            # Extract InstrumentComponent modifications
            ic_modifs = ObsMetadata({ic: mods.pop(ic)
                                    for ic in ic_names if ic in mods})
            # Put remaining modifications back
            attributes_dict['modifications'] = mods
        else:
            ic_modifs = ObsMetadata({})

        # Put shortcuts in the right place:
        if 'serial_number' in attributes_dict:
            x = attributes_dict.pop('serial_number')
            attributes_dict.safe_update(
                {'modifications': {'equipment': {'serial_number': x}}},
                warn_crush=True)
        if 'datalogger_configuration' in attributes_dict:
            x = attributes_dict.pop('datalogger_configuration')
            channel_modifs.safe_update(
                {'*-*': {'datalogger': {'configuration': x}}},
                warn_crush=True)

        # Get main elements
        base_dict = attributes_dict.get_configured_modified_base()
        base_location_code = base_dict.pop('location_code', station_location_code)
        ### 'configuration' is solely informational
        base_configuration = base_dict.pop('configuration', None)
        self.equipment = Equipment(base_dict.pop('equipment', None))
        bcd = base_dict.pop('configuration_description', None)
        if 'configuration_description' in base_dict:
            self.equipment.description += f' [config: {base_dict.pop("configuration_description")}]'
        channels_dict = base_dict.pop('channels', {})
        channel_default = channels_dict.pop('default')

        # Fill in channels
        self.channels = Channels()
        for das_label, ch_attributes in channels_dict.items():
            if not 'location_code' in ch_attributes:
                ch_attributes['location_code'] = base_location_code
            ch_modifs = self._select_channel_modifs(ch_attributes, channel_modifs)
            
            # Extract InstrumentComponent modifiers and update ic_modifiers dict
            ic_names = ('datalogger', 'sensor', 'preamplifier')
            ic_ch_modifs = ic_modifs.copy()
            ic_ch_modifs.safe_update({ic: ObsMetadata(ch_modifs.pop(ic))
                                      for ic in ic_names if ic in ch_modifs})

            # Update dictionairies
            ch_attributes.safe_update(ch_modifs)
            ch_dict = channel_default.copy()
            # There can only be one orientation code
            if 'orientation' in ch_dict and 'orientation' in ch_attributes:
                if not ch_dict['orientation'].keys() == ch_attributes['orientation'].keys():
                    ch_dict['orientation'] = ch_attributes['orientation']
            ch_dict.safe_update(ch_attributes)
            
            # Put start_date and end_date in dictionary
            if not 'start_date' in ch_dict:
                ch_dict['start_date'] = station_start_date
            if not 'end_date' in ch_dict:
                ch_dict['end_date'] = station_end_date

            # Select location
            location = station_locations.get_by_loc_code(ch_dict.pop('location_code'))
            
            # Create channel
            # self.channels.append(Channel(ch_dict, ic_ch_modifs, location,
            #                              self.equipment))
            self.channels.append(Channel(ch_dict, ic_ch_modifs, location,
                                         self.equipment))
        verify_dict_is_empty(base_dict)

    def __repr__(self):
        s = (f'Instrumentation(attributes_dict={self.input["attributes_dict"]}, '
             f'station_locations={self.input["station_locations"]}, '
             f'station_location_code={self.input["station_location_code"]}, '
             f'station_start_date={self.input["station_start_date"]}, '
             f'station_end_date={self.input["station_end_date"]})')
        return s

    def __str__(self, indent=0, n_subclasses=0):
        if n_subclasses < 0:
            return self.__class__.__name__
        kwargs = dict(indent=4, n_subclasses=n_subclasses-1)
        s = 'Instrumentation:\n'
        s += f'    equipment: {self.equipment.__str__(**kwargs)}\n'
        s += f'    channels: {self.channels.__str__(**kwargs)}\n'
        return str_indent(s, indent)

    @staticmethod
    def _select_channel_modifs(chan_dict:dict, channel_modifs:dict):
        """
        Select channel_modifications for the provided chan_dict

        Args:
            chan_dict (dict or :class:`ObsMetadata`): channel attributes.
                Must have 'orientation' and 'location_code' keys
            channel_modifs (dict): channel_modifications, keys are
                id_codes, in order of priority:
                    "<ORT>-<LOC>": orientation_code <ORT> & location_code <LOC>
                    "<ORT>": orientation_code <ORT>
                    "*-<LOC>": location_code <LOC>, all orientation_codes
                    "<ORT>-*": orientation_code <ORT>, all location_codes
                    "*-*": All ``id_codes``
            location_code (str): The location

        Returns:
            (:class:`ObsMetadata``): the selected channel modifications
        """
        for k in ('orientation', 'location_code'):
            if k not in chan_dict:
                raise ValueError(f'chan_dict has no "{k}" key')
        orientation_code = list(chan_dict['orientation'].keys())[0]
        location_code = chan_dict['location_code']
        id_code = orientation_code + "-" + location_code
        
        # Get modifications, classed by specification type
        mods_specific = channel_modifs.get(id_code, channel_modifs.get(id_code[:1], {}))
        mods_orient = channel_modifs.get("*-" + location_code, {})
        mods_loc = channel_modifs.get(orientation_code + "-*", {})
        mods_default = channel_modifs.get("*", channel_modifs.get("*-*", {}))

        # Put modifications in order of priority
        chmods = ObsMetadata(mods_default)
        chmods.safe_update(mods_orient)
        chmods.safe_update(mods_loc)
        chmods.safe_update(mods_specific)
        return chmods



#!/usr/bin/env python
# -*- coding: utf-8 -*-

"""
 Main functions for obsinfo-validate.

 obsinfo strongly follows the hierarchy of StationXML files.

 Module contains both the class ValidateObsinfo and a main entry point for obsinfo-validate
"""
import os
from pathlib import Path
import glob
import re
import sys
from argparse import ArgumentParser

from ..misc.discoveryfiles import Datapath
from ..obsmetadata import ObsMetadata
from ..misc.const import EXIT_SUCCESS
from ..helpers import init_logging

logger = init_logging("validate")

class ValidateObsinfo():
    """
    Methods to validate each level of information files.

    Attributes:
        datapath (:class:`/.Datapath`): Store datapath list(s)
        verbose (bool): Prints progress messages
        remote (bool): Indicates file in command line argument is to be found
            using datapath
        debug (bool): Print more detailed messages.
    """
    # SHOULD THIS BE INIT (looks like Luis took from unittest)
    def setUp(self, verbose=True, remote=False, debug=False):
        """
        Set up status variables according to .obsinforc and command line arguments.

        Args:
            verbose (bool): Print several progress messages
            remote (bool): Find file in command line argument using datapath
            debug (bool): Print more detailed messages and enable traceback
                of exceptions
        Returns: self
        """
        dp = Datapath()
        self.datapath = dp
        self.datapath.infofiles_path = dp.datapath_list
        self.datapath.validate_path = Path(__file__).parent.parent.joinpath('data', 'schemas')

        self.verbose = verbose
        self.remote = remote
        self.debug = debug

        return self

    def validate_single_file(self, info_file, filetype, file_format='yaml',
                             quiet=False, check_schema=False):
        """
        Validate a single obsinfo file.

        Args:
            info_file (str): info file to validate. No assumptions made about
                path.
            filetype (str): the information file type
            file_format (str): the information file format ('json' or 'yaml')
            check_schema (bool): validate schema file as well
        Returns:
            (bool): True if validated, False if not
        """
        if file_format not in ('json', 'yaml'):
            logger.error('file_format ("{file_format}") not in ("json", "yaml")')
            raise ValueError('file_format ("{file_format}") not in ("json", "yaml")')
        try:
            ret = ObsMetadata().validate(info_file,
                                         self.datapath.validate_path,
                                         self.remote,
                                         file_format,
                                         filetype,
                                         self.verbose,
                                         filetype + '.schema.json',
                                         quiet,
                                         check_schema=check_schema)
        except Exception as e:
            print(str(e))
            return False
        if self.verbose:
            self._print_passed(f'{filetype} test for: {info_file}', ret)
        return ret

    def validate_directory(self, dir, forced_filetype=None, drilldown=False,
                           continue_on_fail=False, check_schema=False):
        """
        Validate all information files in a directory.

        Args:
            dir (:class:`Path`): directory where files reside
            forced_filetype (str): file type.  If None, determine it.
            drilldown (bool): drill down through directories beneath the specified one
            continue_on_fail (bool): continue testing if a validation fails
            check_schema (bool): check schema on first use
        Returns:
            false if any file failed
        """
        suffixes = ['.yaml', '.json']
        checked_schemas = []
        if not dir.is_dir():
            raise ValueError(f'"{dir}" is not a directory!')
        if drilldown is True:
            print(f'Validating files in and below directory {dir}')
            if forced_filetype is not None:
                raise ValueError('Cannot specify filetype when drilling down '
                                 'directories')
            files = dir.glob('**/*.*')
        else:
            print(f'Validating files in directory {dir}')
            files = dir.glob('*.*')
        print(f'Will skip files whose suffix is not in "{suffixes}"')
        self_verbose = self.verbose
        filetype = forced_filetype
        any_false = False
        self.verbose = False
        check_this_schema = False
        results = {'passed': 0, 'failed': 0, 'skipped': 0}
        for file in files:
            if file.suffix.lower() not in suffixes:
                print(f'Skipping {str(file.relative_to(dir))}')
                results['skipped'] += 1
                continue
            if forced_filetype==None:
                try:
                    filetype = ObsMetadata.get_information_file_type(str(file))
                except Exception:
                    print(f"Could not find a schema for {str(file.relative_to(dir))}, skipping...")
                    results['skipped'] += 1
                    continue
            print(f"Validating {str(file.relative_to(dir))} against "
                  f"{filetype}.schema.json ... ", end='', flush=True)
            if check_schema is True:
                if filetype in checked_schemas: 
                    check_this_schema = False  
                else:
                    check_this_schema = True   
                    checked_schemas.append(filetype)
            validated = self.validate_single_file(str(file), filetype,
                                                  check_schema=check_this_schema)
            if validated is not True:
                results['failed'] += 1
                print("FAILED")
                any_false = True
                if continue_on_fail is False:
                    print('Quitting on first error')
                    self.verbose = self_verbose
                    return False
            else:
                print("PASSED")
                results['passed'] += 1
        print(80 * '=')
        print('{:d} passed, {:d} failed, {:d} skipped'.format(
            results['passed'], results['failed'], results['skipped']))
        self.verbose = self_verbose
        if any_false is True:
            return False
        return True

    def validate_filters_in_directory(self, dir):
        """
        Validate all information files in filter directory.

        Args:
            dir (str): directory where filter files reside
        """
        if dir.is_dir():
            for file in dir.iterdir():
                self.validate_single_file(file, "filter")

    def validate_stages_in_directory(self, dir):
        """
        Validate all information files in stage directory as given by datapath.

        Args:
            1dir (str): directory where stage files reside
        """
        datalogger_dir_re = re.compile(".*/dataloggers")
        sensor_dir_re = re.compile(".*/sensors")
        exception_re = re.compile(".*/test-with")
        if re.match(datalogger_dir_re, str(dir)) or re.match(sensor_dir_re, str(dir)):
            for file in (Path(dir).joinpath("responses")).iterdir():
                if not file.is_dir() and re.match(exception_re, str(file)) is None:
                    self.validate_single_file(file, "stage")

    def validate_all_components(self):
        """
        Validate all information files in each components
        
        (sensor, preamplifier, datalogger) subdirectory as given by datapath
        """
        components_list = ("sensors", "preamplifiers", "dataloggers")
        for dir in self.datapath.infofiles_path:
            for comp in components_list:
                # includes sensors, preamplifiers and dataloggers
                files_in_validate_dir = Path(dir).joinpath(comp, "*.yaml")
                self.validate_files(files_in_validate_dir, comp)

    def validate_all_filters(self):
        """
        Validate all filter files in datapath/<component>/responses/filters/"
        """
        for dir in self.datapath.infofiles_path:
            # "*rs includes sensors, preamplifiers and dataloggers"
            files_in_validate_dir = Path(dir).joinpath("*rs", "responses",
                                                       "filters", "*.yaml")
            self.validate_files(files_in_validate_dir, "filter")

    def validate_all_stages(self):
        """
        Validate all stage files in datapath/<component>/responses/
        """
        for dir in self.datapath.infofiles_path:
            # "*rs includes sensors, preamplifiers and dataloggers"
            files_in_validate_dir = Path(dir).joinpath("*rs", "responses",
                                                       "*.yaml")
            self.validate_files(files_in_validate_dir, "stage")

    def validate_all_instrumentations(self):
        """
        Validate all instrumentation files in datapath/instrumentation/
        """
        for dir in self.datapath.infofiles_path:
            # includes sensors, preamplifiers and dataloggers
            files_in_validate_dir = Path(dir).joinpath(
                "instrumentation/*.yaml")
            self.validate_files(files_in_validate_dir, "instrumentation")

    def validate_all_networks(self):
        """
        Validate all network files in datapath/network/
        """
        for dir in self.datapath.infofiles_path:
            # includes sensors, preamplifiers and dataloggers
            files_in_validate_dir = Path(dir).joinpath("network/*.yaml")
            self.validate_files(files_in_validate_dir, "network")

    def validate_files(self, files, filetype):
        """
        Validate all files of a given type

        Args:
            files (:class:`Path`): Full paths of files (including wildcards)
            filetype (str): information file type
        """
        assert ObsMetadata.is_valid_type(filetype)

        filelist = glob.glob(str(files))
        for file in filelist:
            self.validate_single_file(file, filetype)

    @staticmethod
    def _print_passed(text, passed):
        if passed:
            print(f'{text}: PASSED')
        else:
            print(f'{text}: FAILED')


def main():
    """
    Entry point for obsinfo-validate.

     1) Sets up status variables from command line arguments.
     2) Validates file according to file type contained in name
     3) Manages all uncaught exceptions
    """
    args = retrieve_arguments()

    val = ValidateObsinfo()
    val.setUp(verbose=args.verbose, remote=args.remote, debug=args.debug)

    filetype = None
    if args.schema:
        if ObsMetadata.is_valid_type(args.schema):
            filetype = args.schema
        else:
            raise ValueError(f'Unknown schema type: {args.schema}')

    if args.input.is_file():
        if filetype is None:
            filetype = ObsMetadata.get_information_file_type(str(args.input))
        if args.verbose:
            print(f'Validating {filetype} file')
        val.validate_single_file(str(args.input), filetype,
                                 check_schema=args.check_schema)
    elif args.input.is_dir():
        if args.schema and args.drilldown:
            raise ValueError("Cannot specify schema for a drill-down "
                             "directory validation")
        val.validate_directory(args.input, filetype, args.drilldown,
                               args.continue_on_fail, check_schema=args.check_schema)

    sys.exit(EXIT_SUCCESS)


def retrieve_arguments():
    """
    Retrieve command line arguments.

    Returns:
        dictionary object with all status variables and information file name.
    """
    # Parse the arguments
    parser = ArgumentParser(
        prog="obsinfo-validate",
        description="Validates obsinfo information files against schema.")

    # flags
    parser.add_argument(
        "-q", "--quiet", action='store_true', default=False,
        help="Quiet operation. Don't print informative messages")
    parser.add_argument(
        "-s", "--schema",
        help="Force validation against the given schema file (sets --check_schema)")
    parser.add_argument(
        "-r", "--remote", action='store_true', default=False,
        help="Search for the input_filename in the DATAPATH repositories")
    parser.add_argument(
        "-d", "--debug", action='store_true', default=False,
        help="Print traceback for exceptions")
    parser.add_argument(
        "--drilldown", action='store_true', default=False,
        help="Drill down through all subdirectories (if a directory "
             "was specified)")
    parser.add_argument(
        "--continue_on_fail", action='store_true', default=False,
        help="Continue validating if a file fails (if a directory was specified)")
    parser.add_argument(
        "--check_schema", action='store_true', default=False,
        help="Check the schema before validating")
    # positional arguments
    parser.add_argument(
        "input", type=str,
        help="Information file or directory to be validated. If a directory, "
             "tests all files in the directory.  If 'DATAPATH', "
             "will test all files in the DATAPATH (sets --drilldown)")

    args = parser.parse_args()
    args.verbose = not args.quiet

    if args.input == 'DATAPATH':
        args.input = Datapath().datapath_list[0]
        args.drilldown = True
        print(f'Validating first DATAPATH dir')

    if args.schema is not None:
        args.check_schema = True
    
    args.input = Path(args.input)
    if not args.input.is_absolute():
        args.input = Path(os.getcwd()).joinpath(args.input).resolve()

    return args


if __name__ == '__main__':
    main()



"""
Application to configure variables used by obsinfo in file .obsinforc
and to copy examples and templates out of the distribution directories
"""
import os
# import sys
import re
# import site
import shutil
import datetime
import platform
from pathlib import Path
import warnings

from argparse import ArgumentParser

from ..misc.const import EXIT_IOERR
from ..version import __version__
from ..helpers import init_logging

logger = init_logging("setup")


def main():
    """
    Entry point to configure variables in file .obsinforc

    and to copy examples and templates out of the distribution directories
    """
    # initialize variables
    version = platform.python_version()
    path_to_package = Path(__file__).parent.parent
    examples_dir = path_to_package.joinpath("_examples/Information_Files")
    templates_dir = path_to_package.joinpath("templates")
    home = Path.home()
    configuration_file = home.joinpath('.obsinforc')

    args = retrieve_arguments()

    validate_arguments(args)

    # create logging directory if it doesn't exist
    logdir = Path().home().joinpath('.obsinfo').resolve()
    if not os.path.isdir(logdir):
        os.mkdir(logdir)

    # copy examples and templates
    if args.dest and not args.no_copy:
        try:
            dest = Path(args.dest).resolve()
            # if dest dir exists rename it with timestamp to preserve it
            if os.path.isdir(dest) or os.path.isfile(dest):
                new_dest = str(dest) + "." + str(datetime.datetime.today())
                os.rename(dest, new_dest)

            shutil.copytree(examples_dir, dest.joinpath("Information_Files"))
            shutil.copytree(templates_dir, dest.joinpath("templates"))
        except NotADirectoryError:
            print("Directory not found")
            exit(EXIT_IOERR)
        except OSError as e:
            print("Operating system error")
            print(e)
            exit(EXIT_IOERR)
    # create configuration file
    output = []

    output.append("gitlab_repository: " + args.gitlab + "\n")
    output.append("gitlab_project: " + args.project + "\n")
    output.append("gitlab_path: " + args.remote_path + "\n")
    pat = re.compile("^[0-9]+\.[0-9]+")
    version = pat.match(__version__)
    branch = "v" + version.group(0) if args.branch else "master"

    output.append('obsinfo_branch: ' + branch + '\n')

    remote_repository = args.gitlab + "/" + args.project + "/" + args.remote_path \
        if not args.no_remote else ""

    s = build_datapath_list(args, remote_repository)
    if s:
        output.append(s + "\n")
    try:
        with open(configuration_file, "w") as fp:
            fp.writelines(output)
    except OSError:
        warnings.warn("Could not create configuration file. "
                      "Fix problem and run obsinfo-setup again\n")
        exit(EXIT_IOERR)


def retrieve_arguments():
    """
    Parse command line arguments

    Returns:
        parse_args (dict): arguments, both status and directory names

    """
    # Parse the arguments
    parser_args = ArgumentParser(prog="obsinfo-setup")

    # flags
    parser_args.add_argument("-x", "--no_examples", action='store_true', default=False,
                             help="Don't import examples, only templates, and "
                                  "remove examples directory from the datapath")
    parser_args.add_argument("-c", "--no_copy", action='store_true', default=False,
                             help="Don't import anything at all, don't create "
                                  "dest directory, which will be removed from datapath")
    parser_args.add_argument("-n", "--no_remote", action='store_true', default=False,
                             help="Install obsinfo without access to a gitlab "
                             "repository.\nMay be needed in some operating "
                             "systems for compatibility")
    parser_args.add_argument("-v", "--invert_datapath", action='store_true', default=False,
                             help="Put remote gitlab repositorey first. "
                                  "All local directories will keep their order")
    parser_args.add_argument("-b", "--branch", action='store_true', default=False,
                             help="Specifies the git branch to use, if not master")

    # optional arguments
    parser_args.add_argument("-d", "--dest", default=None,
                             help="Destination directory for templates and examples.")
    parser_args.add_argument("-g", "--gitlab", default="https://www.gitlab.com",
                             help="Gitlab repository)")
    parser_args.add_argument("-p", "--project", default="resif/obsinfo",
                             help="path to project and the directory where "
                                  "information files lie within the Gitlab "
                                  "repository")
    parser_args.add_argument("-l", "--local_repository", default=None,
                             help="Specify local repository for information "
                                  "files and include it as first or second "
                                  "option in datapath")
    parser_args.add_argument("-w", "--working_directory", default=None,
                             help="Specify working directory for obsinfo and "
                                  "include it as first option in datapath")
    parser_args.add_argument("-P", "--remote_path",
                             default="obsinfo/_examples/Information_Files",
                             help="Specify remote directory under project")
    return parser_args.parse_args()


def validate_arguments(args):
    """
    Validate that directories in command line arguments exist.

    Raise an exception if not.
    Args:
        args (dict of str): directory with command line argument, but status
            and directories
    Raises:
        OSError, ValueError
    """

    if args.dest:
        if not Path(args.dest):
            raise OSError(ENOENT, "Path for example and template destination "
                                  "directory is an illegal path")
    else:
        default_information_files = ""

    if args.local_repository:
        if not Path(args.local_repository):
            raise OSError(ENOENT, "Path for local repository for information "
                                  "files is an illegal path")

    if args.working_directory:
        if not Path(args.working_directory):
            raise OSError(ENOENT, "Path for obsinfo working directory is an "
                                  "illegal path")

    if not args.no_remote:
        if not Path(args.gitlab).as_uri:
            raise ValueError('Gitlab repository must be specified as a legal '
                             'URI unless "-n" is on')

    if not args.no_remote:
        if not Path(args.project):
            raise ValueError('Project must be specified as a legal URI unless '
                             '"-n" is on')


def build_datapath_list(args, remote_repository):
    """
    Build datapath as a list of directories to search
    
    Similar to the PATH environment variable.

    #. Add working_directory as first option if it exists
    #. Next add local repository first if it exists (wll be first if no
       working directory specified)
    #. Finally add destination, where standard examples lie.
       This won't be used if -x is on
    #. Final directory will be the remote one unless -v (invert) is specified
    """

    default_information_files = str(Path(args.dest).joinpath("Information_Files").resolve()) if args.dest else ""
    remote_first = remote_repository + ", " if args.invert_datapath else ""
    remote_last = remote_repository if not args.invert_datapath else ""

    s = "datapath: ["
    s += remote_first
    s += args.working_directory + ", " if args.working_directory else ""
    s += args.local_repository + ", " if args.local_repository else ""
    s += default_information_files + ", " if default_information_files and not args.no_examples else ""
    s += remote_last if remote_last else ""
    s += "]"

    if not remote_last:
        s.replace(", ]", "]")

    return s



"""
Channel, Instrument and Operator classes
"""
# Standard library modules
import warnings
import logging

# Non-standard modules
from obspy.core.inventory.channel import Channel as obspy_Channel
from obspy.core.inventory.util import Comment

# obsinfo modules
from ..obsmetadata import ObsMetadata
from ..helpers import (Location, OIDate, str_indent, verify_dict_is_empty,
                       Comments, ExternalReferences, Identifiers, ObsinfoClassList)
from .orientation import Orientation
from .instrument import Instrument

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Channels(ObsinfoClassList):
    """
    A list of Channel objects
    """
    def __init__(self, channels_list=None):
        """
        Args:
            channels_list: (list of :class:`Instrumentations`):
        """
        if channels_list is None:
            super().__init__([], Channel)
        elif isinstance(channels_list, list):
            super().__init__(channels_list, Channel)
        else:
            raise TypeError('channels_list is neither a list nor None')


class Channel(object):
    """
    Corresponds to StationXML/obspy Channel plus channel code

    Attributes:
        das_channel (:class:`ObsMetadata`): represents a channel with defaults
            incorporated
        location (:class:`Location`): location for this channel
        start_date (str): inherited from Station
        end_date (str): inherited from Station
        instrument (:class:`Instrument`): a sensor, a datalogger and an
            optional preamplifier
        orientation (:class:`Orientation`):
        comments (list of str)
    """

    def __init__(self, attributes_dict, ic_modifs: dict, location, equipment):
        """
        Constructor

        Args:
            attributes_dict (dict or :class:`ObsMetadata`): channel attributes
            ic_modifs (dict or :class:`ObsMetadata`): modifications to pass
                through to InstrumentComponents
            location (:class:`Location`): channel location
            equipment (:class:`Equipment`): channel equipment
        """
        # For __repr__
        self.input = {
                      'attributes_dict':   {} if not attributes_dict else '<ObsMetaData>',
                      'ic_modifs':   {} if not ic_modifs else '<ObsMetaData>',
                      'location': location,
                      'equipment': '<obsinfo.Equipment>'
                     }

        orientation = attributes_dict.pop('orientation', None)
        o_code = self._get_orientation_code(orientation)
        self.equipment = equipment
        self.location = location
        if 'location_code' in attributes_dict:
            if not self.location.code == attributes_dict['location_code']:
                raise ValueError('location.code != channel["location_code"] '
                                 f'({self.location.code}!={attributes_dict["location_code"]})')
            del attributes_dict['location_code']
        self.start_date = OIDate(attributes_dict.pop('start_date'))
        self.end_date = OIDate(attributes_dict.pop('end_date'))
        self.comments = Comments(attributes_dict.pop("comments", None))
        self.comments += Comments.from_extras(attributes_dict.pop("extras", None))

        self.orientation = Orientation(orientation) # ,
                                       # self.instrument.polarity)
        self.source_id = attributes_dict.pop('source_id', None)
        self.identifiers = Identifiers(attributes_dict.pop('identifiers', None))
        self.restricted_status = attributes_dict.pop('restricted_status', None)
        self.external_references = ExternalReferences(attributes_dict.pop('external_references', None))

        self.instrument = Instrument(attributes_dict, ic_modifs)

        verify_dict_is_empty(attributes_dict)

    def __repr__(self):
        s =  f'Channel(default_attributes={self.input["default_attributes"]}\n'
        s += f'        ch_attributes={self.input["ch_attributes"]},\n'
        s += f'        ic_modifs={self.input["modifs"]},\n'
        s += f'        location={self.input["location"]}\n'
        s += f'        equipment={self.input["equipment"]}'
        s += f'       )'
        return s

    def __str__(self, indent=0, n_subclasses=0):
        """
        Args:
            indent (int): number of extra characters to indent lines by
            n_subclasses (int): number of levels of subclass to print out
        """
        if n_subclasses < 0:
            return self.__class__.__name__

        kwargs = {'indent': 4, 'n_subclasses': n_subclasses-1}
        s = f'Channel {self.seed_code}:\n'
        s += f'    location code: {self.location.code}\n'
        # Force print of at least location.__str__() and orientation.__str__()
        s += f'    location: {self.location.__str__(**kwargs)}\n'
        s += f'    orientation: {self.orientation.__str__(**kwargs)}\n'
        s += f'    start_date: {self.start_date}\n'
        s += f'    end_date: {self.end_date}\n'
        s += f'    equipment: {self.equipment.__str__(**kwargs)}\n'
        s += f'    comments: {self.comments}\n'
        s += f'    equipment: {self.instrument.__str__(**kwargs)}\n'

        return str_indent(s, indent)

    @property
    def seed_code(self):
        """
        This is equivalent to channel code for self.instrument.sample_rate
        """
        return self.channel_code(self.instrument.sample_rate)

    def channel_code(self, sample_rate):
        """
        Return channel code for a given sample rate.

        Validates instrument and orientation codes according to FDSN
        specifications (for instruments, just the length). Channel codes
        specified by user are indicative and are refined using actual sample
        rate.

        :param sample_rate: instrumentation sampling rate (sps)
        :type sample_rate: float
        """
        inst_code = self.instrument.sensor.seed_instrument_code

        if len(inst_code) != 1:
            msg = f'Instrument code "{inst_code}" is not a single letter'
            warnings.warn(msg)
            logger.error(msg)
            raise ValueError(msg)
        if self.orientation.code not in ["X", "Y", "Z", "1", "2", "3",
                                         "H", "F"]:
            msg = f'Orientation code "{self.orientation.code}" is not a '\
                  'valid letter'
            warnings.warn(msg)
            logger.error(msg)
            raise ValueError(msg)

        band_code = self._get_band_code(sample_rate)
        return (band_code + inst_code + self.orientation.code)

    def to_obspy(self):
        """
         Create obspy Channel object

         Returns:
            (~class:`obspy.core.inventory.channel.Channel`)
        """
        channel = obspy_Channel(
            self.seed_code,
            self.location.code,
            latitude=self.location.latitude.to_obspy(),
            longitude=self.location.longitude.to_obspy(),
            elevation=self.location.elevation.to_obspy(),
            depth=self.location.depth_m,
            water_level=self.location.water_level,
            azimuth=self.orientation.azimuth.to_obspy(),
            dip=self.orientation.dip.to_obspy(),
            types=['CONTINUOUS', 'GEOPHYSICAL'],
            external_references=self.external_references,
            sample_rate=self.instrument.sample_rate,
            clock_drift_in_seconds_per_sample=1 / (1e8 * float(self.instrument.sample_rate)),
            sensor=self.instrument.sensor.equipment.to_obspy(),
            pre_amplifier=self.instrument.preamplifier.equipment.to_obspy()
                          if self.instrument.preamplifier else None,
            data_logger=self.instrument.datalogger.equipment.to_obspy(),
            equipments=[self.equipment.to_obspy()],
            response=self.instrument.to_obspy(),
            comments=self.comments.to_obspy(),
            start_date=self.start_date.to_obspy(),
            end_date=self.end_date.to_obspy(),
            restricted_status=self.restricted_status,
            identifiers=self.identifiers.to_obspy(),
            source_id=self.source_id,
            sample_rate_ratio_number_samples=None,
            sample_rate_ratio_number_seconds=None,
            description=None,
            calibration_units=None,
            calibration_units_description=None,
            storage_format=None,
            alternate_code=None,
            historical_code=None,
            data_availability=None)
        return channel

    def _channel_id(self, orientation_code):
        """
        Uniquely identify channel through orientation and location code

        format: {orientation}-{location}
        Args:
            orientation_code (str): single-letter orientation code
        :returns: channel code
        """
        return orientation_code + "-" + self.location.code

    def _get_orientation_code(self, orientation_dict):
        """
        Get the orientation code from a dict that may include azimuth and dip
        """
        # There should be only one key
        return list(orientation_dict.keys())[0]

    def _get_band_code(self, sample_rate):
        """
        Return the channel band code corresponding to a sample rate

        :param sample_rate: sample rate (sps)
        :type sample_rate: float
        """
        bbc = self.instrument.sensor.seed_band_base_code
        if len(bbc) != 1:
            msg = f'Band base code "{bbc}" is not a single letter'
            warnings.warn(msg)
            logger.error(msg)
            raise ValueError(msg)
        if bbc in "FCHBMLVUWRPTQ":
            if sample_rate >= 1000:
                return "F"
            elif sample_rate >= 250:
                return "C"
            elif sample_rate >= 80:
                return "H"
            elif sample_rate >= 10:
                return "B"
            elif sample_rate > 1:
                return "M"
            elif sample_rate > 0.3:
                return "L"
            elif sample_rate >= 0.1:
                return "V"
            elif sample_rate >= 0.01:
                return "U"
            elif sample_rate >= 0.001:
                return "W"
            elif sample_rate >= 0.0001:
                return "R"
            elif sample_rate >= 0.00001:
                return "P"
            elif sample_rate >= 0.000001:
                return "T"
            else:
                return "Q"
        elif bbc in "GDES":
            if sample_rate >= 1000:
                return "G"
            elif sample_rate >= 250:
                return "D"
            elif sample_rate >= 80:
                return "E"
            elif sample_rate >= 10:
                return "S"
            else:
                msg = "Short period sensor sample rate < 10 sps"
                warnings.warn(msg)
                logger.warning(msg)
                return "S"  # Return a code anyway
        else:
            msg = f'Unknown band base code: "{bbc}"'
            warnings.warn(msg)
            logger.error(msg)
            raise TypeError(msg)

    @staticmethod
    def _get_comments(comments=[], extras={}):
        extras_list = [str(k) + ": " + str(v) for k, v in (extras).items()]
        if extras_list:
            comments.extend([f'Extra attribute: {{{e}}}' for e in extras_list])
        return comments



"""
Analog Filter (subclass of PolesZeros)
"""
# Standard library modules
import logging

from .poles_zeros import PolesZeros
from ...helpers import str_indent

logger = logging.getLogger("obsinfo")


class Analog(PolesZeros):
    """
    Analog Filter (Flat PolesZeros filter)
    """
    def __init__(self, attributes_dict, stage_id=-1):
        """
        Constructor

        Args:
            attributes_dict (dict or :class:`.ObsMetadata`): filter attributes
            stage_id (int): id of corresponding stage. Used for reporting only
        """
        logger.debug(f'in {self.__class__.__name__}.__init__()')
        attributes_dict["transfer_function_type"] = "LAPLACE (RADIANS/SECOND)"
        attributes_dict["poles"] = []
        attributes_dict["zeros"] = []
        if not "normalization_factor" in attributes_dict:
            attributes_dict["normalization_factor"] = 1.0
        # if not "normalization_frequency" in attributes_dict:
        #     attributes_dict["normalization_frequency"] = 1.
        super().__init__(attributes_dict, stage_id)
        self._validate_empty_attributes_dict(attributes_dict)

        if not (self.normalization_frequency and self.normalization_factor):
            self.normalization_factor = 1.0

    def __repr__(self):
        return 'Analog(attributes_dict, stage_id)'

    def __str__(self, indent=0, n_subclasses=0):
        if n_subclasses < 0:
            return f'{type(self)}'
        s = super().__str__()
        return str_indent(s, indent)



"""
obsinfo information file routines, contained in superclass ObsMetadata for
generality
"""
# Standard library modules
import json
from pathlib import Path, PurePath
from urllib.parse import urlparse
from urllib.parse import unquote
import logging
import time

# Non-standard modules
import jsonschema
import yaml

# Local modules
from ..misc import yamlref
from ..misc.yamlref import JsonLoader
from ..misc.remoteGitLab import gitLabFile
from ..misc.discoveryfiles import Datapath

logger = logging.getLogger("obsinfo")

overwrite_symbol = '^'      # Used by safe_update
root_symbol = "#"
VALID_FORMATS = ["JSON", "YAML"]
DEFAULT_SCHEMA_PATH = Path(__file__).parent.parent.joinpath('data', 'schemas')
VALID_TYPES = [Path(f.stem).stem
               for f in DEFAULT_SCHEMA_PATH.glob('*.schema.json')]


class ObsMetadata(dict):
    def __init__(self, *args, **kwargs):
        """
        Constructor, create a dict subclass object

        Args:
            args (*list): positional arguments
            kwargs (**dict): keyword arguments
        """
        super().__init__(*args, **kwargs)
        self._convert_to_obsmetadata()

    def list_valid_types():
        """
        Returns a list of valid information file types

        :returns: list of valid information file types
        """
        return VALID_TYPES

    @staticmethod
    def is_valid_type(type):
        """
        Returns true if input is a valid schema type
        """
        return type in VALID_TYPES

    def validate(self, info_filename, schemas_path=None, remote=False,
                 file_format=None, file_type=None, verbose=False,
                 schema_filename=None, quiet=False, dp=None,
                 check_schema=False):
        """
        Validates a YAML or JSON file against schema

        Args:
            info_filename (str or path-like): name of information file to
                validate
            schemas_path (str or :class:`Path`): path to schema files.
                If None, uses DEFAULT_SCHEMA_PATH
            remote (bool): whether to search for info_filename in a remote
                repository
            file_format (str): "JSON" or "YAML"
            file_type (str): type of info file: "network", "station",
                "instrumentation", "datalogger", "preamplifier", "sensor",
                "stage", "filter"
            verbose (bool): Print progression of validation steps with
                filenames
            schema_file (str): name (without path) of schema file
            quiet (bool): No output at all
            dp (Datapath): datapath for information files.  If None,
                defaults to values stored in .obsinforc
            check_schema (bool): validate schema file as well

        If file_type and/or format are not provided, tries to figure them out
        from the info_filename, which should end with "*{FILETYPE}.{FORMAT}*
        """
        if quiet:
            verbose = False

        # Get schema filename
        if not schema_filename:
            if not file_type:
                file_type = ObsMetadata.get_information_file_type(info_filename)
            schema_filename = file_type + '.schema.json'

        # Get infofile instance and schema
        if dp is None:
            dp = Datapath()
        instance = self.read_info_file(info_filename, dp, remote, False,
                                       file_format, verbose)
        return self.validate_infodict(instance, schema_filename, verbose,
                                      quiet, schemas_path, check_schema)

    def validate_infodict(self, info_dict, schema_filename, verbose=False,
                          quiet=False, schemas_path=None, check_schema=True):
        """
        Validates an infofile dict file against schema

        Args:
            info_dict (dict or :class:`ObsMetadata``): information to
                validate
            verbose (bool): Print progression of validation steps with
                filenames
            schema_filename (str): name (without path) of schema file
            schemas_path (str or :class:`Path`): path to schema files.
                If None, uses DEFAULT_SCHEMA_PATH
            check_schema (bool): validate schema file as well
        """
        # Get schema filename
        if ".schema.json" not in schema_filename:
            schema_filename += '.schema.json'

        instance = ObsMetadata(info_dict)

        schema, schema_fullpath = self._read_schema_file(schema_filename,
                                                         schemas_path)
        if schema is False:
            logger.error('Could not read schema file {} at {}'.format(
                schema_filename, schemas_path))
            return False
        return self._report_errors(instance, schema, schema_fullpath, verbose,
                                   quiet, check_schema)

    @staticmethod
    def check_schema(schema, verbose, quiet=False):
        """This is really slow! (45s for instrumentation, 240s for subnetwork!)

        Returns:
            result (bool): True if schema checks out, false if not
        """
        tic = time.perf_counter()
        msg = "\tTesting schema ..."
        logger.info(msg)
        if verbose:
            print(msg, end="")

        try:
            jsonschema.Draft7Validator.check_schema(schema)
        except jsonschema.ValidationError as e:
            logger.exception("SCHEMA ERROR: " + e.message)
            return False
        logger.info("OK")
        if verbose:
            print("OK")
        if quiet is not True:
            print(f'schema test took {time.perf_counter()-tic:.1f} seconds')
        return True

    def get_information_file_format(filename):
        """
        Determines if the information file is in JSON or YAML format.

         Assumes that the filename is "*.{FORMAT}*

        Args:
            filename (str): filename to determine the type of
        Returns:
            file_format
        Raises:
            (ValueError): on unknown format
        """
        suffix = PurePath(filename).suffix
        file_format = suffix[1:].upper()
        if file_format in VALID_FORMATS:
            return file_format
        msg = f"Unknown file_format: {file_format}"
        logger.error(msg)
        raise ValueError(msg)

    def get_information_file_type(filename):
        """
        Determines the type of a file.

        Assumes that the filename is "*.{TYPE}.{SOMETHING}*

        Args:
            filename (str): filename to determine the type of
        Returns:
            (str): file type
        Raises:
            ValueError
        """
        stem = PurePath(filename).stem
        suffix = PurePath(stem).suffix
        type = suffix[1:]
        if type in VALID_TYPES:
            return type
        msg = f"File '{filename}' is of unknown type: {type}"
        logger.error(msg)
        raise ValueError(msg)

    def read_json_yaml(filename, file_format=None):
        """
        Reads a JSON or YAML file. Does NOT use jsonReference  DEPRECATED.

        DEPRECATED. Not being used by any obsinfo method or function. Kept
        for compatibility

        Args:
            filename (str): filename
            file_format (str): "YAML" or "JSON"
        Returns:
            (dict):  JSON or YAML parsed information files
        Raises:
            (JSONDecodeError): problem with JSON read
            (FileNotFoundError): file not found
            (IOError): File input/output erre
        """
        if not file_format:
            # Also validates that format is legal and exits if not
            file_format = ObsMetadata.get_information_file_format(filename)

        with open(filename, "r") as f:
            if file_format == "YAML":
                try:
                    element = yaml.safe_load(f)
                except Exception:
                    msg = f"Error loading YAML file: {filename}"
                    logger.exception(msg)
                    raise
            else:
                try:
                    element = json.load(f)
                except json.JSONDecodeError as e:
                    msg = ("JSONDecodeError: Error loading JSON file: "
                           f"{filename}: {str(e)}")
                    logger.exception(msg)
                    raise
                except Exception:
                    msg = f"Error loading JSON file: {filename}"
                    logger.exception(msg)
                    raise
        return element

    def read_json_yaml_ref_datapath(filename, datapath, file_format=None):
        """
        Reads a JSON or YAML file using jsonReference using OBSINFO_DATAPATH

        Args:
            filename (str): filename
            datapath (:class:`.Datapath`): list of directories to search
                for info files
            file_format(str): "YAML" or "JSON"
        Returns:
            (dict):  JSON or YAML parsed information files
        Raises:
            (JSONDecodeError): problem with JSON read
            (FileNotFoundError): file not found
            (IOError): File input/output erre
        """

        if not file_format:
            file_format = ObsMetadata.get_information_file_format(filename)

        bu = unquote(filename)

        if gitLabFile.isRemote(bu):
            base_uri = unquote(urlparse(bu).path)
            loader = JsonLoader()
            jsonstr = loader.get_json_or_yaml(bu, base_uri=base_uri,
                                              datapath=datapath)
            return yamlref.loads(jsonstr, base_uri=base_uri, datapath=datapath)
        else:
            base_uri = Path(bu).as_uri()
            try:
                with open(unquote(filename), "r") as f:
                    return yamlref.load(f, base_uri=base_uri,
                                        datapath=datapath)
            except FileNotFoundError:
                msg = f'File not found: {filename}'
                logger.exception(msg)
                raise
            except (IOError, OSError):
                msg = f'Input/Output error with file: {filename}'
                logger.exception(msg)
                raise

    def read_json_yaml_ref(filename, datapath, file_format=None):
        """
        Reads a JSON or YAML file using jsonReference

        Like read_json_yaml_ref, but does not look for files in
        OBSINFO_DATAPATH
        $ref within the data files without absolute or relative path will be
        still looked for in OBSINFO_DATAPATH

        Args:
            filename (str): filename
            datapath (:class:`.Datapath`): object to store list of
            directories to search info files. Used as a dummy.
            file_format (str): "YAML" or "JSON"
        Returns:
            (dict): JSON or YAML parsed information files
        Raises:
            (JSONDecodeError): problem with JSON read
            (FileNotFoundError): file not found
            (IOError): File input/output erre
        """
        if not file_format:
            file_format = ObsMetadata.get_information_file_format(filename)

        bu = unquote(filename)
        base_uri = Path(bu).as_uri()

        try:
            with open(filename, "r") as f:
                a = yamlref.load(f, base_uri=base_uri, datapath=datapath)
            return a
        except FileNotFoundError as error:
            msg = f'File not found: {error.args}'
            logger.exception(msg)
            raise
        except (IOError, OSError) as error:
            msg = f'Input/Output error with file: {error.args}'
            logger.exception(msg)
            raise

    @staticmethod
    def read_info_file(filename, datapath, remote=False, validate=True,
                       file_format=None, verbose=False,  quiet=False):
        """
        Reads an information file

        Args:
            filename (str): filename
            datapath (:class:`.Datapath`): stores list of directories to
                search info files
            validate (bool): validate before reading
            remote (bool): whether to use absolute/relative path locally
                or OBSINFO_DATAPATH
            file_format (str):"YAML" or "JSON"
            verbose (bool): say more
            quiet (bool): say nothing 

        Returns:
            (:class:`ObsMetadata`): JSON or YAML parsed info files
        """
        if quiet is True:
            verbose = False
        if validate:
            file_type = ObsMetadata.get_information_file_type(filename)
            msg = f'Validating {file_type} file: {filename}'
            logger.info(msg)
            if verbose:
                print(msg)
            ObsMetadata().validate(str(filename), DEFAULT_SCHEMA_PATH,
                                   remote=remote, verbose=verbose, quiet=quiet)
        else:
            msg = f"Reading {filename}"
            logger.info(msg)
            if verbose:
                print(msg)

        if remote:
            return ObsMetadata.read_json_yaml_ref_datapath(filename, datapath,
                                                           file_format)
        else:
            return ObsMetadata.read_json_yaml_ref(filename, datapath,
                                                  file_format)

    def get_configured_modified_base(self, higher_modifs={}):
        """
        Return a fully configured and modified base_dict
        
        Handles ``base``, ``modifications`` and ``configuration``
        fields in self and higher_modifs.  Doest not handle any of
        these subelements in ``modifications`` or ``configuration``: they
        simply get added into ``base`` using the ``safe_update()`` method.
        
        Values in higher-modifs outrank those in self.  Modifications outrank
        configurations.  Uses safe_update() to only change specified elements.
        
        Args:
            self (:class:`ObsMetadata`): base-configuration-modification
                dictionary.  Must have "base", can have "configuration" and
                "modification" AND NOTHING ELSE.
            higher_modifs (dict or :class:`ObsMetadata`): modifications
                dictionary.  Can have "base", "configuration" and/or
                "modification" AND NOTHING ELSE
        Returns:
                base_dict (:class:`ObsMetadata): fully configured and modified
                    attribute dictionary
        Raises:
            ValueError: if self or higher_modifs contain keys other than "base",
                "configuration" and/or "modification"
        """
        a = self.copy()  
        base_dict = self.__class__(a.pop('base', {}))
        configuration = base_dict.pop('configuration_default', None)
        if 'configuration' in a:
            configuration = a.pop('configuration')
        modifs = self.__class__(a.pop('modifications', {}))
        if len(list(a.keys())) > 0:
            raise ValueError('base-configuration-modification dict '
                             f'has leftover keys: {a}')

        # Update with higher-level modifications
        b = self.__class__(higher_modifs.copy())
        if 'base' in b:
            base_dict = self.__class__(b.pop("base"))
            configuration = base_dict.pop('configuration_default', None)
            modifs = ObsMetadata({})
        if 'configuration' in b:
            configuration = b.pop("configuration")
        if "modifications" in b:
            high_modifs = b.pop("modifications")
            # modifications can also specify BASE and CONFIGURATION
            if 'base' in high_modifs:
                base_dict = self.__class__(high_modifs.pop("base"))
                configuration = base_dict.pop('configuration_default', None)
                modifs = ObsMetadata({})
            if 'configuration' in high_modifs:
                configuration = high_modifs.pop('configuration')
            modifs.safe_update(high_modifs)
        if len(list(b.keys())) > 0:
            raise ValueError('higher_level base-configuration-modification '
                             f'dict had leftover elements: {b}')

        # Configure, then modify
        configs = base_dict.pop("configurations", None)
        if configuration is not None:
            if configs is None:
                raise ValueError(f"'{configuration}' configuration was requested"
                                 ', but no configurations were specified')
            if configuration not in configs:
                raise ValueError(
                    f"Requested configuration ('{configuration}') doesn't "
                    f"match specified configurations: {list(configs.keys())}")
            base_dict.safe_update(configs[configuration])
            base_dict['configuration'] = configuration
        base_dict.safe_update(modifs)

        return base_dict
        
    def safe_update(self, update_dict, allow_overwrite=True, warn_crush=False):
        """
        Update that only changes explicitly specfied fields

        Drills recursively through dicts inside the dict, only changing fields
        which are specified in update_dict.  Lists are completely replaced, 
        however, to avoid ambiguity.

        Args:
            update_dict (dict or :class:`ObsMetadata`): dictionary containing
                fields to update
            allow_overwrite (bool): allow a field that was originally a dict
                to be overwritten by a field that is not a dict.  Same for lists.
            warn_crush (bool): write out a warning when a value is replaced
        """
        if not isinstance(update_dict, dict):
            logger.error('update_dict is not a dict')
            raise TypeError('update_dict is not a dict')
        if not isinstance(update_dict, self.__class__):
            update_dict = self.__class__(update_dict)
        for key, value in update_dict.items():
            # Change any dict into ObsMetadata
            if isinstance(value, dict) and not isinstance(value, self.__class__):
                value = self.__class__(value)
            if key[0] == overwrite_symbol:
                key=key[1:]
                if key in self:
                    logger.info('Overwrite ordered for key {key}')
                self[key] = value
                continue  # Go to next loop iteration
            if key not in self:  # Add new key and its value
                self[key] = value
                continue  # Go to next loop iteration
            # Key exists in self and not forced overwrite
            if isinstance(self[key], dict):  # If original item is a dict
                # if value is also a dictionary, update it
                if isinstance(value,  dict):
                    # if replacement value is a dictionary, recurse
                    self[key].safe_update(value,
                                          allow_overwrite=allow_overwrite,
                                          warn_crush=warn_crush)
                else:
                    # if replacement value is not a dictionary
                    if allow_overwrite:  # replace & warn
                        self[key] = value
                        logger.warning(f'dict field "{key}" '
                                       'was replaced by a non-dict')
                    else:  # reject & warn
                        logger.warning(
                            f'replacement field "{key}" was not inserted '
                            'into original because original was a dict '
                            'but replacement was not')
            elif isinstance(self[key], list):  # If original item is a list
                if isinstance(value,  list):   # If replacement is also a list
                    # Replace the list
                    self[key] = value
                    logger.debug(f'"{key}": {len(self[key])}-element list replaced by {len(value)}-element list')
                    # Recurse on contents
                    # replacer = []
                    # for item, holder in zip(value, self[key]):
                    #     if isinstance(item, (dict, ObsMetadata)):
                    #         holder = ObsMetadata(holder)  # SHOULDN'T BE NECESSARY!
                    #         holder.safe_update(self.__class__(item),
                    #                            allow_overwrite=allow_overwrite,
                    #                            warn_crush=warn_crush)
                    #     elif isinstance(item, list):
                    #         raise ValueError('does not handle lists in lists')
                    #     else:
                    #         holder = item
                    #     replacer.append(holder)
                    # self[key] = replacer
                else:
                    # if replacement value is not a list
                    if allow_overwrite:  # replace & warn
                        self[key] = value
                        msg = f'field "{key}" was a list, replaced by a non-list'
                        logger.warning(msg)
                    else:  # reject & warn
                        msg = (f'replacement field "{key}" not inserted '
                               'into original because original was a list '
                               'but replacement was not')
                        logger.warning(msg)
            else:
                # Replace existing others
                if key in self and warn_crush is True:
                    msg = f'replacing self["{key}"]: was {self[key]}, now {value}'
                    logger.warning(msg)
                self[key] = value

    def copy(self):
        return ObsMetadata(super().copy())

    def _convert_to_obsmetadata(self):
        """
        Make all contained dictionaries objects of :class: `.ObsMetadata`
        """
        for key, value in self.items():
            if isinstance(value, dict):
                self[key] = self.__class__(value)
                self[key]._convert_to_obsmetadata()
            elif isinstance(value, list):
                for x in value:
                    if isinstance(x, dict):
                        x = self.__class__(x)
                        x._convert_to_obsmetadata()

    @staticmethod
    def _read_schema_file(base_file, schemas_path=None):
        if schemas_path is None:
            schemas_path = DEFAULT_SCHEMA_PATH
        schema_fullpath = PurePath(schemas_path) / base_file
        base_uri = unquote(PurePath(schema_fullpath).as_uri())
        schema_datapath = Datapath(schemas_path)
        try:
            with open(schema_fullpath, "r") as f:
                try:
                    s = f.read()
                except Exception:
                    logger.exception()
                    return False, False
                try:
                    schema = yamlref.loads(
                        s, base_uri=base_uri, jsonschema=True,
                        datapath=schema_datapath, recursive=True)
                except json.decoder.JSONDecodeError as e:
                    msg = ("JSONDecodeError: Error loading JSON schema "
                           f"file: {schema_fullpath}")
                    logger.exception(msg)
                    logger.error(str(e))
                    return False, False
                except BaseException as e:
                    msg = "{}: Error loading JSON schema file: {}".format(
                        type(e), schema_fullpath)
                    logger.exception(msg)
                    logger.error(e)
                    return False, False
        except FileNotFoundError:
            msg = f'File not found: {schema_fullpath}'
            logger.exception(msg)
            raise FileNotFoundError(msg)
        except (IOError, OSError):
            msg = f'Input/Output error with file: {schema_fullpath}'
            logger.exception(msg)
            raise
        return schema, schema_fullpath

    @staticmethod
    def _report_errors(instance, schema, schema_fullpath=Path(""),
                       verbose=False, quiet=False, check_schema=True):
        """
        Lazily report all errors in the instance vis-a-vis the schema

        ASSUMES DRAFT7 SCHEMA (I couldn't get it to work otherwise)
        """
        msg = f"schema =   {schema_fullpath.name}"
        logger.info(msg)
        if verbose:
            print(msg)
        if check_schema:
            if not ObsMetadata.check_schema(schema, verbose, quiet):
                return False

        msg = "\tTesting instance ..."
        logger.info(msg)
        if verbose:
            print(msg, end="")

        v = jsonschema.Draft7Validator(schema)

        if not v.is_valid(instance):
            # if not quiet:
            #     print("")
            errors = sorted(v.iter_errors(instance), key=lambda e: e.path)
            for error in errors:
                err_path = ''.join([f"['{e}']" for e in error.path])
                msg = f"{err_path}: {error.message} \tFAILED"
                # if not quiet:
                #     print("\t\t" + msg)
                logger.error(msg)  # errors get printed to console
            return False
        else:
            if verbose:
                print("OK")
            logger.info("OK")
            return True



"""
Filter class and subclasses
"""
# Standard library modules
import logging

from .poles_zeros import PolesZeros
from .analog import Analog
from .digital import Digital
from .AD_conversion import ADConversion
from .FIR import FIR
from .response_list import ResponseList
from .coefficients import Coefficients
from .polynomial import Polynomial
from ...obsmetadata import ObsMetadata

logger = logging.getLogger("obsinfo")


class Filter(object):
    """
    Filter is a gateway to the other filter classes
    """
    @staticmethod
    def construct(attributes_dict, stage_sequence_number, stage_name,
                  sensitivity_gain_frequency=1.):
        """
        Constructs an appropriate Filter subclass from an attributes_dict

        Args:
            attributes_dict (dict or list of dicts): information file
                dict for stages
            stage_sequence_number (int): sequence_number of corresponding stage.
            stage_name (str): name of corresponding stage. Used for reporting only
            sensitivity_gain_frequency (float): frequency at which gain was
                specified.
                Used for PoleZeros Normalization factor/frequency
        Returns:
            (:class:`.Filter`): object of the adequate filter subclass
        Raises:
            (TypeError): if filter type is not valid
        """
        logger.debug('in Filter.construct()')
        if attributes_dict is None:
            msg = "No attributes in filter"
            logger.error(msg)
            raise TypeError(msg)
        if not isinstance(attributes_dict, ObsMetadata):
            attributes_dict = ObsMetadata(attributes_dict)

        stage_description = f'#{stage_sequence_number}'
        if stage_name is not None:
                stage_description += f' ("{stage_name}")'

        if "type" not in attributes_dict:
            msg = f'No "type" specified for filter in stage {stage_description}'
            logger.error(msg)
            raise TypeError(msg)
        else:
            args = (attributes_dict, stage_description)
            filter_type = attributes_dict.get('type', None)
            if filter_type == 'PolesZeros':
                attributes_dict['sensitivity_gain_frequency'] = sensitivity_gain_frequency
                obj = PolesZeros(*args)
            elif filter_type == 'FIR':
                obj = FIR(*args)
            elif filter_type == 'Coefficients':
                obj = Coefficients(*args)
            elif filter_type == 'ResponseList':
                obj = ResponseList(*args)
            elif filter_type == 'ADCONVERSION':
                obj = ADConversion(*args)
            elif filter_type == 'ANALOG':
                attributes_dict['sensitivity_gain_frequency'] = sensitivity_gain_frequency
                obj = Analog(*args)
            elif filter_type == 'DIGITAL':
                obj = Digital(*args)
            elif filter_type == 'Polynomial':
                obj = Polynomial(*args)
            else:
                msg = (f'Unknown Filter type: "{filter_type}" in '
                       'stage #{stage_id}')
                logger.error(msg)
                raise TypeError(msg)
        return obj



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Functions to print obsinfo objects
"""


class PrintObs(object):
    """
    Collection of methods to print obsinfo objects at different levels of depth.
    All methods are static.

    Attributes:
        *None*
    """

    @staticmethod
    def print_component(obj, level="all"):
        """
        Prints comoponent information and continues if level is not "component" or "channel".

        If level is not "channel" detailed equipment information is not printed.

        Args:
            level (str): determines to which level the *obsinfo* object will
                be printed
        """
        if not obj:
            return
        print(obj)
        if level == "channel":
            return
        print(obj.equipment)

    @staticmethod
    def print_instrumentation(obj, level="all"):
        """
        Prints instrumentation information and continues if level is not "instrumentation".

        If level is  "response" or "all" response stages information will be printed (if it exists).
        Recall at this point all the response is gathered in ``instrument.stages.sgates``
        If level is "all" or "filter" (i.e. not "response") all information will be printed (only filter is left at this point...)

        Args:
            level (str): determines to which level the *obsinfo* object will
                be printed
        """
        print(obj)
        print("Instrument:\n")
        print(obj.equipment)
        for k in obj.channels:
            print(k)
            if level != "instrumentation":
                print(k.instrument)
                PrintObs.print_component(k.instrument.sensor, level)
                PrintObs.print_component(k.instrument.preamplifier, level)
                PrintObs.print_component(k.instrument.datalogger, level)

                if level in ["response", "all", "filter"] and k.instrument.stages:
                    for s in k.instrument.stages:
                        print(s)
                        if level in ["all", "filter"]:
                            print(s.filter)

    @staticmethod
    def print_station(obj, level="all"):
        """
        Prints station information and continues if level is not "station".

        Args:
            level (str): determines to which level the *obsinfo* object will
                be printed
        """
        print(obj)
        print(obj.locations)
        for k, v in obj.locations.items():
            print(f'Location {k}: {v}')
        if level != "station":
            PrintObs.print_instrumentation(obj.instrumentation, level)

    @staticmethod
    def print_network(obj, level="all"):
        """
        Prints network information and continues if level is not "network".

        Args:
            level (str): determines to which level the *obsinfo* object will
                be printed
        """
        print(obj)
        print(obj.operator)
        if level != "network":
            for v in obj.stations:
                PrintObs.print_station(v, level)



"""
Identifiers Class, used in Networks, Channels and Stations
"""
# Standard library modules
import warnings
import logging
import json

from obspy.core.inventory.util import ExternalReference as obspy_extref

from .functions import str_indent
from .person import Persons
from .oi_date import OIDate
from .obsinfo_class_list import ObsinfoClassList

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Identifiers(ObsinfoClassList):
    """
    A list of Identifier objects
    """
    def __init__(self, attributes_list):
        """
            Args:
                attributes_list: (list of dict or str): list
                    of external references read from YAML/JSON file
        """
        if attributes_list is None:
            super().__init__([], Identifier)
        else:
            super().__init__(attributes_list, Identifier)


class Identifier(object):
    """
    A type to document persistent identifiers. Must provide a scheme (prefix)
    """
    def __init__(self, uri):
        if ':' not in uri:
            raise ValueError(f'would-be Identifier "{uri}" has no scheme (prefix + ":")')
        self.uri = uri

    def __str__(self, indent=0, n_subclasses=0):
        """ Always a one-liner """
        return(f'{self.__class__.__name__}: {self.uri}')

    def to_obspy(self):
        """
        obspy handles identifiers as single strings
        """
        return self.uri



"""
Person class
"""
# Standard library modules
import warnings
import logging

from obspy.core.inventory.util import Person as obspy_Person

from .phone import Phones
from .functions import str_indent, str_list_str
from .obsinfo_class_list import ObsinfoClassList

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Persons(ObsinfoClassList):
    """
    A list of Person objects
    """
    def __init__(self, seq):
        """
        Args:
            seq: (list of Person or None): list of attribute dicts
                describing a Person
        """
        super().__init__(seq, Person)


class Person(object):
    """
    Equivalent of StationXMl Person

    Attributes:
        names (list of str): Name(s) written out normally
        agencies (list of str): agencies worked for
        emails (list of str): emails
        phones (list of Phone): person's telephone number(s)
    """

    def __init__(self, attributes_dict):
        """
        Args:
            attributes_dict (dict or :class:`.ObsMetadata`):
                operator information
        """
        if not attributes_dict:
            raise ValueError('No attributes dict!')

        self.names = attributes_dict['names']                  # required
        self.agencies = attributes_dict.get('agencies', [])
        self.emails = attributes_dict.get('emails', [])
        self.phones = Phones(attributes_dict.get('phones', []))

    def repr(self):
        args = [f'names={self.names}']
        if len(self.agencies) > 0:
            args.append(f'agencies={self.agencies}')
        if len(self.emails) > 0:
            args.append(f'emails={self.emails}')
        if len(self.phones) > 0:
            args.append(f'phones={self.phones}')
        return 'Person(' + ", ".join(args) + ")"

    def __str__(self, indent=0, n_subclasses=0):
        if n_subclasses < 0:
            return f'{type(self)}'
        kwargs = dict(indent=indent, n_subclasses=n_subclasses-1)
        s = f'{self.__class__.__name__}:\n'
        s += f'    names: {str_list_str(self.names, **kwargs)}\n'
        s += f'    agencies: {str_list_str(self.agencies, **kwargs)}\n'
        s += f'    emails: {str_list_str(self.emails, **kwargs)}\n'
        s += f'    phones: {self.phones.__str__(**kwargs)}'
        return str_indent(s, indent)

    def to_obspy(self):
        return obspy_Person(names=self.names,
                            agencies=self.agencies,
                            emails=self.emails,
                            phones=self.phones.to_obspy())



"""
InstrumentComponent class and subclasses Sensor, Preamplifier, Datalogger.
Equipment class
"""
# Standard library modules
import logging

# Non-standard modules

# obsinfo
from .stages import Stages
from .equipment import Equipment
from ..obsmetadata import (ObsMetadata)
from ..helpers import str_indent, str_list_str, verify_dict_is_empty

logger = logging.getLogger("obsinfo")


class InstrumentComponent(object):
    """
    InstrumentComponent class. Superclass of all component classes.
    No obspy/StationXML equivalent, because they only specify the whole
    sensor+amplifier+datalogger system

    Attributes:
        equipment (:class:`Equipment`)
        stages (:class:`Stages`)
        obspy_equipment (:class:`obspy_Equipment`)
        configuration_description (str): description of configuration to be added
            to equipment description
    """

    def __init__(self, attributes_dict, higher_modifs={}):
        """
        Creator.

        Args:
            attributes_dict (dict or :class:`ObsMetadata`): InstrumentComponent
                attributes
            higher_modifs (dict or :class:`ObsMetadata`): modifications
                inherited from instrumentation
        """
        if not attributes_dict:
            msg = 'No attributes_dict'
            logger.error(msg)
            raise TypeError(msg)
        if not isinstance(attributes_dict, ObsMetadata):
            attributes_dict = ObsMetadata(attributes_dict)

        # Remove elements to be processed in Stages()
        higher_stage_modifications = ObsMetadata(attributes_dict.pop('stage_modifications', {}))
        if 'stage_modifications' in higher_modifs:
            higher_stage_modifications.safe_update(higher_modifs.pop('stage_modifications'))

        # Put shortcuts in the right place:
        if 'serial_number' in attributes_dict:
            attributes_dict.safe_update(
                {'modifications':
                    {'equipment':
                        {'serial_number': attributes_dict.pop('serial_number')}
                }},
                warn_crush=True)
        if 'serial_number' in higher_modifs:
            higher_modifs.safe_update(
                {'modifications':
                    {'equipment':
                        {'serial_number': higher_modifs.pop('serial_number')}
                }},
                warn_crush=True)

        # Get base-configured-modified element
        base_dict = attributes_dict.get_configured_modified_base(higher_modifs)
        
        if 'notes' in base_dict:
            del base_dict['notes']

        self.equipment = Equipment(base_dict.pop('equipment', None))
        self.configuration = base_dict.pop('configuration', None)
        self.configuration_description = base_dict.pop('configuration_description', self.configuration)
        stage_modifications = ObsMetadata(base_dict.pop('stage_modifications', {}))
        stage_modifications.safe_update(higher_stage_modifications, warn_crush=True)
        self.stages = Stages(base_dict.pop('stages', []), {'stage_modifications': stage_modifications},
                             base_dict.get('correction', None))
        self.equipment.calibration_dates += self.stages.calibration_dates
        logger.info('Instrument_Component has {len(self.stages)} stages')
        self._error_check(self.stages)
        if self.configuration_description is not None:
            self.equipment.description += ' [config: {}]'.format(
                self.configuration_description)
        # self.obspy_equipment = self.equipment.to_obspy()
        self.base_dict = base_dict  # leftovers for specific components

    def __repr__(self):
        s = ''
        if self.equipment.description:
            s += f', description="{self.equipment.description}"'
        if self.stages is not None:
            s += f'Response stages: {len(self.stages)}'
        return s

    def __str__(self, indent=0, n_subclasses=0):
        """
        Information common to all instrument_components, to be appended
        Args:
            indent (int): number of extra characters to indent lines by
            n_subclasses (int): number of levels of subclass to print out
        """
        kwargs = {'indent': 4, 'n_subclasses': n_subclasses-1}
        s = f'\n    equipment: {self.equipment.__str__(**kwargs)}'
        if self.stages is not None:
            s += f'\n    stages: {self.stages.__str__(**kwargs)}'
        return s

    def _error_check(self, stages):
        """Some of these checks seem redundant"""
        if stages is None:
            msg = f'stages is None {type(self)}'
            logger.warning(msg)
        elif not stages:
            msg = f'No response stages in {type(self)}'
            logger.error(msg)
            raise TypeError(msg)
        elif len(stages) == 0:
            msg = f'len(stages) == 0 in {type(self)}'
            logger.error(msg)
            raise TypeError(msg)


class Sensor(InstrumentComponent):
    """
    Sensor Instrument Component. No obspy equivalent

    Attributes:
        equipment (:class:`.Equipment`): Equipment information
        seed_band_base_code (str (len 1)): SEED base code ("B" or "S")
            indicating instrument band.  Must be modified by
            obsinfo to correspond to output sample rate. Actual SEED base
            code is determined by FDSN standard <http://docs.fdsn.org/
            projects/source-identifiers/en/v1.0/channel-codes.html>`
        seed_instrument code(str (len 1)): SEED instrument code, determined
            by `FDSN standard <http://docs.fdsn.org/projects/source-
            identifiers/en/v1.0/channel-codes.html>`
        stages (:class:`Stages`): channel modifications
            inherited from station
    """
    def __init__(self, attributes_dict, higher_modifs={}):
        """
        Create Sensor instance from an attributes_dict

        Args:
            attributes_dict (dict or :class:`ObsMetadata`): InstrumentComponent
                attributes
            higher_modifs (dict or :class:`ObsMetadata`): modifications
                inherited from instrumentation
        """
        logger.info('Creating Sensor()')
        super().__init__(attributes_dict, higher_modifs)
        seed_dict = self.base_dict.pop('seed_codes', {})
        self.seed_band_base_code = seed_dict.get('band_base', None)
        self.seed_instrument_code = seed_dict.get('instrument', None)
        verify_dict_is_empty(self.base_dict)
        del self.base_dict

    def __str__(self, indent=0, n_subclasses=0):
        if n_subclasses < 0:
            return f'{type(self)}'
        s = f'{self.__class__.__name__}:'
        s += f"\n    band code: {self.seed_band_base_code}"
        s += f"\n    instrument code: {self.seed_instrument_code}"
        s += super().__str__(indent, n_subclasses)
        return str_indent(s, indent)


class Datalogger(InstrumentComponent):
    """
    Datalogger Instrument Component.

    Obspy ``Datalogger` only contains elements of :class:`Equipment`, rest is
    in Response

    Attributes:
        equipment (:class:`Equipment`): equipment attributes
        stages (:class:`Stages`): channel modifications
            inherited from station
        sample_rate (float): sample rate of given configuration. Checked
            against actual sample rate
        correction (float or None): the delay correction of the
            component. If a float, it is applied to the last stage and the
            other stage corrections are set to 0.  If None, each stage's
            correction is set equal to its delay
    """
    def __init__(self, attributes_dict, higher_modifs={}):
        """
        Args:
            attributes_dict (dict or :class:`ObsMetadata`): InstrumentComponent
                attributes
            higher_modifs (dict or :class:`ObsMetadata`): modifications
                inherited from instrumentation
        Returns:
            (:class:`Datalogger`)
        """
        logger.info('Creating Datalogger()')
        super().__init__(attributes_dict, higher_modifs)
        self.sample_rate = self.base_dict.pop('sample_rate', None)
        self.correction = self.base_dict.pop('correction', 0)
        verify_dict_is_empty(self.base_dict)
        del self.base_dict

    def __str__(self, indent=0, n_subclasses=0):
        if n_subclasses < 0:
            return f'{type(self)}'
        s = f'{self.__class__.__name__}:'
        s += f"\n    sample_rate: {self.sample_rate}"
        s += f"\n    correction: {self.correction}"
        s += super().__str__(indent, n_subclasses)
        return str_indent(s, indent)


class Preamplifier(InstrumentComponent):
    """
    Preamplifier Instrument Component. No obspy equivalent

      Attributes:
        equipment (:class:`Equipment`): Equipment information
        stages (:class:`Stages`): channel modifications inherited from station
        configuration_description (str): the configuration description that was
            selected, added to equipment description
    """
    def __init__(self, attributes_dict, higher_modifs={}):
        """
        Args:
            attributes_dict (dict or :class:`ObsMetadata`): InstrumentComponent
                attributes
            higher_modifs (dict or :class:`ObsMetadata`): modifications
                inherited from instrumentation
        """
        logger.info('Creating Preamplifier()')
        if not attributes_dict:
            return None   # It is acceptable to have no preamplifier
        super().__init__(attributes_dict, higher_modifs)
        verify_dict_is_empty(self.base_dict)
        del self.base_dict

    def __str__(self, indent=0, n_subclasses=0):
        if n_subclasses < 0:
            return f'{type(self)}'
        s = f'{self.__class__.__name__}:'
        s += super().__str__(indent, n_subclasses)
        return str_indent(s, indent)

    def __repr__(self):
        s = super().__str__(indent, n_subclasses)
        return str_indent(s, indent)



#!/usr/bin/env python
# -*- coding: utf-8 -*-
"""
Test StationXML creation
"""
from pathlib import Path
import glob
import unittest
import inspect
# from pprint import pprint
import xml.etree.ElementTree as ET
from CompareXMLTree import XmlTree
import warnings

from obsinfo.main.makeStationXML import main as make_StationXML
from obsinfo.misc.discoveryfiles import (Datapath)


warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
verbose = False

     
class StationXMLTest(unittest.TestCase):
    """
    Class of test methods for StationXML creation
    """

    """
    Test suite for obsinfo operations.
    """
    def setUp(self):
        self.testing_path = Path(__file__).parent / "data_main"
        self.infofiles_path = (Path(__file__).resolve().parents[1] /
                                "_examples" / 'Information_Files')

    def test_makeSTATIONXML(self):
        """
        Test STATIONXML creation.
        """
        for fname in ["SPOBS.INSU-IPGP.subnetwork.yaml",
                      "BBOBS.INSU-IPGP.subnetwork.yaml"]:
            net_file = str(self.infofiles_path / "subnetwork" / fname)
            make_StationXML([net_file, '--quiet'],
                             Datapath(str(self.infofiles_path)))

            compare = XmlTree()
            # excluded elements
            excludes = ["Created", "Real", "Imaginary", "Numerator",
                        "CreationDate", "Description", "Module"]
            excludes_attributes = ["startDate", "endDate"]
            excludes = [compare.add_ns(x) for x in excludes]

            for stxml in glob.glob("*.xml"):
                xml1 = ET.parse(stxml)
                xml2 = ET.parse(self.testing_path / "StationXML" / stxml)
                self.assertTrue(compare.xml_compare(
                    compare.getroot(xml1), compare.getroot(xml2),
                    excludes=excludes,
                    excludes_attributes=excludes_attributes))
                Path(stxml).unlink()


def suite():
    return unittest.makeSuite(StationXMLTest, 'test')


if __name__ == '__main__':
    unittest.main(defaultTest='suite')




# import sys
# import os
# import re
from pathlib import Path
from urllib import parse as urlparse
# from urllib.parse import urljoin
import warnings
# import yaml
import logging

# obsinfo modules
from ..misc.remoteGitLab import gitLabFile
from .configuration import ObsinfoConfiguration

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")

# First create all the various objects. These strongly follow the
# hierarchy of StationXML files.


class Datapath(object):
    """
    Class to discover where information files are stored.

    Attributes:
        datapath_list (list of str): directories where information files will
            be searched, in sequence
        infofiles_path (list of str): same as datapath_list, used by validate,
            kept for compatibility
        validate_path (str): one unique path to validation schemas
    """

    def __init__(self, datapath=None):
        """
        Args:
            datapath (list of str): directories where information files will be
                searched for. THIS OVERRIDES THE ~.obsinforc FILE!
        """
        if datapath is not None:
            # warnings.warn('You are hand-setting the DATAPATH, will not read'
            #               ' ~/.obsinforc')
            if isinstance(datapath, str):
                datapath = [datapath]
            elif isinstance(datapath, Path):
                datapath = [str(datapath)]
            else:
                assert isinstance(datapath, list)
                for x in datapath:
                    if isinstance(datapath, Path):
                        x = str(x)
                    else:
                        assert isinstance(datapath, str)
        else:
            datapath = ObsinfoConfiguration.Instance().obsinfo_datapath

        if not datapath:
            warnings.warn("OBSINFO_DATAPATH not set, defaulting to working directory")
            self.datapath_list = ["./"]  # Use current directory as default.
        else:
            self.datapath_list = datapath

        self.infofiles_path = None
        self.validate_path = None

    def __str__(self):
        s = f'datapath_list: {self.datapath_list}\n'
        s += f'infofiles_path: {self.infofiles_path}\n'
        s += f'validate_path: {self.validate_path}'
        return s

    def build_datapath(self, file):
        """
        Create list of directories which may have data or schemas

        1) If the file path is absolute, return the file itself.
        2) If path starts by ./ or ../ complete to an absolute path using working directory
        3) If the file has no prefix discover whether file exists in the datapath list.
           Use the first found file.

        Args:
            file (str or path): filename of file to be found
        Returns:
            found file as string
        Raises:
            FileNotFoundError
        """

        file, frag = urlparse.urldefrag(file)

        filestr = file if isinstance(file, str) else str(file)

        if not isinstance(file, Path):
            file = Path(file)

        if file.is_absolute():
            return filestr

        elif filestr[0:3] == "../" or filestr[0:2] == "./":  # if path is absolute or relative to cwd:
            home = Path.cwd()
            self.datapath_list = str(Path.joinpath(home, file).resolve())
            return self.add_frag(self.datapath_list, frag)

        for dir in self.datapath_list:

            if gitLabFile.isRemote(dir):
                # This is done to avoid funny behaviour by pathlib
                slash = "/" if dir[-1] != "/" else ""
                fn = dir + slash + filestr

                if gitLabFile.get_gitlab_file(str(fn), False):  # Check remote repository
                    self.datapath = fn
                    return Datapath.add_frag(self.datapath, frag)  # Don't forget to add frag back!
            else:
                fn = (Path(dir).resolve()).joinpath(file)

                if fn.is_file():  # Check local repository
                    self.datapath = str(fn)
                    return Datapath.add_frag(self.datapath, frag)  # Don't forget to add frag back!

        raise FileNotFoundError(file)

    @staticmethod
    def add_frag(path, frag):
        """
        Add the path and the frag to restore a partial or complete uri

        Args:
            path (str): path portion of uri, possibly with other elements but
                without frag
            frag (str): fragment portion of uri
        Returns:
            (str)L path with frag
        """
        return path + ("#" + frag if frag else "")



"""
InstrumentComponent class and subclasses Sensor, Preamplifier, Datalogger.
Equipment class
"""
# Standard library modules
import warnings
import logging

# Non-standard modules
from obspy.core.inventory.util import Equipment as obspy_Equipment
from obspy.core.utcdatetime import UTCDateTime

# obsinfo
from ..obsmetadata import ObsMetadata
from ..helpers import OIDate, OIDates, str_indent, str_list_str

warnings.simplefilter("once")
warnings.filterwarnings("ignore", category=DeprecationWarning)
logger = logging.getLogger("obsinfo")


class Equipment(obspy_Equipment):
    """
    Equipment.

    Equivalent to :class: obspy.core.inventory.util.Equipment

    Attributes:
        type (str):
        channel_modif (str):
        selected_config (str):
        description (str):
        manufacturer (str):
        model (str):
        vendor (str):
        serial_number (str):
        installation_date (str in date format):
        removal_date (str in date format):
        calibration_dates (str in date format):
        resource_id (str):
        obspy_equipment (class `obspy.core.inventory.equipmentEquipment`)`
    """

    def __init__(self, attributes_dict):
        """
        Constructor

        Args:
            attributes_dict (dict or :class:`ObsMetadata`): attributes of
                component
        """
        attributes_dict = ObsMetadata(attributes_dict)
        self.type = attributes_dict.get('type', None)
        self.description = attributes_dict.get('description', None)
        self.model = attributes_dict.get('model', None)
        self.manufacturer = attributes_dict.get('manufacturer', None)
        self.vendor = attributes_dict.get('vendor', None)
        self.serial_number = attributes_dict.get('serial_number', None)
        self.resource_id = attributes_dict.get('resource_id', None)
        self.installation_date = OIDate(attributes_dict.get('installation_date', None))
        self.removal_date = OIDate(attributes_dict.get('removal_date', None))
        self.calibration_dates = OIDates(attributes_dict.get('calibration_dates', None))
        # equip = self.to_obspy()
        # self.obspy_equipment = equip[0] if isinstance(equip, tuple) else equip

    def __repr__(self):
        args = []
        if self.type:
            args.append(f"'type': '{self.type}'")
        if self.description:
            args.append(f"'description': '{self.description}'")
        if self.manufacturer:
            args.append(f"'manufacturer': '{self.manufacturer}'")
        if self.model:
            args.append(f"'model': '{self.model}'")
        if self.vendor:
            args.append(f"'vendor': '{self.vendor}'")
        if self.serial_number:
            args.append(f"'serial_number': '{self.serial_number}'")
        if self.installation_date.date is not None:
            args.append(f"'installation_date': {self.model}")
        if self.removal_date.data is not None:
            args.append(f"'removal_date': {self.removal_date}")
        if self.calibration_dates:
            args.append(f"'calibration_dates': {self.calibration_dates}")
        return 'Equipment({' + ", ".join(args) + '})'

    def __str__(self, indent=0, n_subclasses=0):
        if n_subclasses < 0:
            return f"{type(self)}"
        s = 'Equipment:\n'
        s += f"    type: {self.type}\n"
        s += f"    description: {self.description}\n"
        s += f"    model: {self.model}"
        if self.manufacturer is not None:
            s += f"\n    manufacturer: {self.manufacturer}"
        if self.vendor is not None:
            s += f"\n    vendor: {self.vendor}"
        if self.serial_number is not None:
            s += f"\n    serial_number: {self.serial_number}"
        if self.resource_id is not None:
            s += f"\n    resource_id: {self.resource_id}"
        if self.installation_date.date is not None:
            s += f"\n    installation_date: {self.installation_date}"
        if self.removal_date.date is not None:
            s += f"\n    removal_date: {self.removal_date}"
        if self.calibration_dates:
            s += f"\n    calibration_dates: {self.calibration_dates.__str__(indent=4, n_subclasses=n_subclasses-1)}"
        return str_indent(s, indent)

    def to_obspy(self):
        """
        Convert an equipment (including the equipment description in
        components) to its obspy object

        Returns:
            (:class:`obspy.core.invertory.util.Equipment`)
        """
        return obspy_Equipment(
            type=self.type,
            description=self.description,
            manufacturer=self.manufacturer,
            vendor=self.vendor,
            model=self.model,
            serial_number=self.serial_number,
            installation_date=self.installation_date.to_obspy(),
            removal_date=self.removal_date.to_obspy(),
            calibration_dates=self.calibration_dates.to_obspy(),
            resource_id=self.resource_id)



"""
stores dates and converts to UTCDateTime
"""
# Standard library modules
import logging
import re

# Non-standard modules
from obspy.core import UTCDateTime

from .obsinfo_class_list import ObsinfoClassList

logger = logging.getLogger("obsinfo")


class OIDates(ObsinfoClassList):
    """
    A list of OIDate objects
    """
    def __init__(self, seq):
        """
        Args:
            seq: (list of OIDate or None): list of attribute dicts
                describing an OIDate
        """
        super().__init__(seq, OIDate)


class OIDate(object):
    """
    Store dates before converting to UTCDateTime

    Attributes:
        date
    """
    def __init__(self, datestr):
        """
        Create object and assign attributes from attributes_dict.

        Args:
            attributes_dict (dict or :class:`ObsMetadata`): dict with
                relevangt keys
        """
        self.date = self.validated_date(datestr)

    def __str__(self, *args, **kwargs):
        if self.date is None:
            return('None')
        return(self.date)

    def __repr__(self):
        return f"OIDate('{self.date}')"

    def __eq__(self, other):
        if not isinstance(other, OIDate):
            return False
        return self.date == other.date

    def to_obspy(self):
        """
        Return UTCDateTime object:
        """
        if self.date is None:
            return None
        return UTCDateTime(self.date)

    @staticmethod
    def validated_date(str_date):
        """
        Reformats an individual date string

        Uses regular expressions to match known dates, either in UTC date
        format or in UTC date and time format.
        The separator can be either "/" or "-"

        Args:
            date (str): a date in a given format
        Returns:
            (str): a reformatted date as string or None if no value
        """
        if str_date is None or str_date == int(0):
            # 0 is sometimes the default, the epoch date, 1/1/1970
            return None

        try:
            dt = UTCDateTime(str_date)
        except Exception as e:
            logger.error(f"Unrecognized date: {str_date}")
            return None
        regexp_date_UTC = re.compile(r"^[0-9]{4}[-\/][0-1]{0,1}[0-9][-\/][0-3]{0,1}[0-9]")
        regexp_date_and_time_UTC = re.compile(r"^[0-9]{4}[-\/][0-1]{0,1}[0-9][-\/][0-3]{0,1}[0-9]T[0-2][0-9]:[0-6][0-9]:{0,1}[0-6]{0,1}[0-9]{0,1}Z{0,1}")
        if not ((re.match(regexp_date_UTC, str_date) or re.match(regexp_date_and_time_UTC, str_date))):
            if dt.hours==0 and dt.minutes==0 and dt.seconds==0 and dt.microseconds==0:
                str_date = dt.strftime('%Y-%m-%d')
            else:
                str_date = dt.isoformat()
        return str_date

    @staticmethod
    def validated_dates(dates):
        """
        Convert list of dates to a standard format

        Args:
            dates (list): dates as strings
        Returns:
            (list): formatted dates as strings
        """
        if dates is None or not isinstance(dates, list):
            return []
        else:
            return [OIDate.validated_date(dt) for dt in dates]
